"""
Here several method and implementation for peakdetector.

  * method = "threshold_and_sum"
     very fast, no spatial information
    2 implementations:
      * numpy
      * opencl
  * method = "spatiotemporal_extrema"
     detect local spatiotemporal extrema
     quite slow
     2 implementations:
       * numpy
       * opencl


"""
import time

import numpy as np

import sklearn.metrics.pairwise

#~ from pyacq.core.stream.ringbuffer import RingBuffer
from .tools import FifoBuffer

from .cltools import HAVE_PYOPENCL, OpenCL_Helper
if HAVE_PYOPENCL:
    import pyopencl
    mf = pyopencl.mem_flags

try:
    import numba
    HAVE_NUMBA = True
    from .numba_tools import numba_get_mask_spatiotemporal_peaks
except ImportError:
    HAVE_NUMBA = False





def detect_peaks_in_chunk(sig, n_span, thresh, peak_sign, spatial_smooth_kernel=None):
    sum_rectified = make_sum_rectified(sig, thresh, peak_sign, spatial_smooth_kernel)
    mask_peaks = detect_peaks_in_rectified(sum_rectified, n_span, thresh, peak_sign)
    time_ind_peaks,  = np.nonzero(mask_peaks)
    time_ind_peaks += n_span
    return time_ind_peaks



class BasePeakDetector:
    offline_function = None

    def __init__(self, sample_rate, nb_channel, chunksize, dtype, geometry):
        self.sample_rate = sample_rate
        self.nb_channel = nb_channel
        self.chunksize = chunksize
        self.dtype = np.dtype(dtype)
        self.geometry = geometry # 2D array (nb_channel, 2 or 3) 

    def change_params(self, peak_sign=None, relative_threshold=None,
                                            peak_span_ms=None, peak_span=None,
                                            smooth_radius_um=None,
                                            adjacency_radius_um=None,
                                            cl_platform_index=None, # only for CL
                                            cl_device_index=None):
        self.peak_sign = peak_sign
        self.relative_threshold = relative_threshold
        
        #~ print('peak_span_ms', peak_span_ms)
        if peak_span_ms is None:
            # kept for compatibility with previous version
            assert peak_span is not None
            peak_span_ms = peak_span * 1000.
        
        self.peak_span_ms = peak_span_ms
        
        self.adjacency_radius_um = adjacency_radius_um
        
        self.n_span = int(self.sample_rate * self.peak_span_ms / 1000.)//2
        #~ print('self.n_span', self.n_span)
        self.n_span = max(1, self.n_span)
        #~ print('self.n_span', self.n_span)
        
        self.smooth_radius_um = smooth_radius_um

        if self.smooth_radius_um is None or self.smooth_radius_um <= 0.:
            self.spatial_smooth_kernel = None
        else:
            d = sklearn.metrics.pairwise.euclidean_distances(self.geometry)
            self.spatial_smooth_kernel = np.exp(-d/self.smooth_radius_um)
            # make it sparse
            self.spatial_smooth_kernel[self.spatial_smooth_kernel<0.01] = 0.
        
        
        self.cl_platform_index = cl_platform_index
        self.cl_device_index = cl_device_index
    

    def process_buffer(self, data):
        # used for offline processing when parralisation  is possible
        raise(NotImplmentedError)
    
    def process_buffer_stream(self, pos, newbuf):
        raise(NotImplmentedError)
    
    def initialize_stream(self):
        # must be for each new segment when index 
        # start back
        raise(NotImplmentedError)
    
    #~ def _after_params(self):
        #~ raise(NotImplmentedError)



def make_sum_rectified(sig, thresh, peak_sign, spatial_smooth_kernel):
    if spatial_smooth_kernel is None:
        sig = sig.copy()
    else:
        sig = np.dot(sig, spatial_smooth_kernel)
    
    
    if peak_sign == '+':
        sig[sig<thresh] = 0.
    else:
        sig[sig>-thresh] = 0.
    
    if sig.shape[1]>1:
        sum_rectified = np.sum(sig, axis=1)
    else:
        sum_rectified = sig[:,0]
    
    return sum_rectified
    

def detect_peaks_in_rectified(sig_rectified, n_span, thresh, peak_sign):
    sig_center = sig_rectified[n_span:-n_span]
    if peak_sign == '+':
        mask_peaks = sig_center>thresh
        for i in range(n_span):
            mask_peaks &= sig_center>sig_rectified[i:i+sig_center.size]
            mask_peaks &= sig_center>=sig_rectified[n_span+i+1:n_span+i+1+sig_center.size]
    elif peak_sign == '-':
        mask_peaks = sig_center<-thresh
        for i in range(n_span):
            mask_peaks &= sig_center<sig_rectified[i:i+sig_center.size]
            mask_peaks &= sig_center<=sig_rectified[n_span+i+1:n_span+i+1+sig_center.size]
    return mask_peaks
    
    #~ time_ind_peaks,  = np.nonzero(mask_peaks)
    #~ time_ind_peaks += n_span
    #~ return time_ind_peaks



class PeakDetectorGlobalNumpy(BasePeakDetector):
    def process_buffer(self, data):
        sig_rectified = make_sum_rectified(data, self.relative_threshold, self.peak_sign, self.spatial_smooth_kernel)
        
        mask_peaks = detect_peaks_in_rectified(sig_rectified, self.n_span, self.relative_threshold, self.peak_sign)
        time_ind_peaks,  = np.nonzero(mask_peaks)
        time_ind_peaks += self.n_span
        chan_ind_peaks = None # not in this method
        peak_val_peaks = None  # not in this method
        
        return time_ind_peaks, chan_ind_peaks, peak_val_peaks
        


    def process_buffer_stream(self, pos, newbuf):
        # this is used by catalogue constructor
        # here the fifo is only the rectified sum
        
        sum_rectified = make_sum_rectified(newbuf, self.relative_threshold, self.peak_sign, self.spatial_smooth_kernel)
        self.fifo_sum_rectified.new_chunk(sum_rectified, pos)
        
        #~ if pos-(newbuf.shape[0]+2*self.n_span)<0:
            # the very first buffer is sacrified because of peak span
            #~ return None, None
        
        #~ sig = self.ring_sum.get_data(pos-(newbuf.shape[0]+2*k), pos)
        sig_rectified = self.fifo_sum_rectified.get_data(pos-(newbuf.shape[0]+2*self.n_span), pos)
        
        mask_peaks = detect_peaks_in_rectified(sig_rectified, self.n_span, self.relative_threshold, self.peak_sign)
        time_ind_peaks,  = np.nonzero(mask_peaks)
        time_ind_peaks += self.n_span
        
        chan_ind_peaks = None # not in this method
        
        if time_ind_peaks.size>0:
            time_ind_peaks = time_ind_peaks + pos - newbuf.shape[0] -2*self.n_span
            return time_ind_peaks, chan_ind_peaks, None

        return None, None, None
    
    def get_mask_peaks_in_chunk(self, fifo_residuals):
        # this is used by peeler which handle externaly
        # a fifo residual
        sum_rectified = make_sum_rectified(fifo_residuals, self.relative_threshold, self.peak_sign, self.spatial_smooth_kernel)
        mask_peaks = detect_peaks_in_rectified(sum_rectified, self.n_span, self.relative_threshold, self.peak_sign)
        return mask_peaks        
    
    def change_params(self, **kargs):
        BasePeakDetector.change_params(self,  **kargs)
        self.fifo_sum_rectified = FifoBuffer((self.chunksize*2,), self.dtype)
    
    def initialize_stream(self):
        self.fifo_sum_rectified = FifoBuffer((self.chunksize*2,), self.dtype)

class PeakDetectorGlobalOpenCL(BasePeakDetector, OpenCL_Helper):
    def process_buffer(self, data):
        #TODO
        raise(NotImplmentedError)
    
    def process_buffer_stream(self, pos, newbuf):
        if newbuf.shape[0] <self.chunksize:
            newbuf2 = np.zeros((self.chunksize, self.nb_channel), dtype=self.dtype)
            newbuf2[-newbuf.shape[0]:, :] = newbuf
            newbuf = newbuf2

        if not newbuf.flags['C_CONTIGUOUS']:
            newbuf = newbuf.copy()

        pyopencl.enqueue_copy(self.queue,  self.sigs_cl, newbuf)
        event = self.kern_detect_peaks(self.queue,  self.global_size, self.local_size,
                                self.sigs_cl, self.ring_sum_cl, self.peak_mask_cl)
        event.wait()
        
        #~ if pos-(newbuf.shape[0]+2*self.n_span)<0:
            # the very first buffer is sacrified because of peak span
            #~ return None, None
        
        pyopencl.enqueue_copy(self.queue,  self.peak_mask, self.peak_mask_cl)
        time_ind_peaks,  = np.nonzero(self.peak_mask)
        
        if time_ind_peaks.size>0:
            time_ind_peaks += pos - self.chunksize - self.n_span
            chan_ind_peaks = None# not in this method
            return time_ind_peaks, chan_ind_peaks, None
        
        return None, None, None
    
    def initialize_stream(self):
        pass
    
    def change_params(self, cl_platform_index=None, cl_device_index=None, **kargs):
        BasePeakDetector.change_params(self,  **kargs)
        OpenCL_Helper.initialize_opencl(self, cl_platform_index=cl_platform_index, cl_device_index=cl_device_index)

        if self.chunksize > self.max_wg_size:
            n = int(np.ceil(self.chunksize / self.max_wg_size))
            self.global_size =  (self.max_wg_size * n, )
            self.local_size = (self.max_wg_size,)
        else:
            self.global_size = (self.chunksize, )
            self.local_size = (self.chunksize, )


        chunksize2=self.chunksize+2*self.n_span
        
        self.sum_rectified = np.zeros((self.chunksize), dtype=self.dtype)
        self.peak_mask = np.zeros((self.chunksize), dtype='uint8')
        ring_sum = np.zeros((chunksize2), dtype=self.dtype)
        
        #GPU buffers
        self.sigs_cl = pyopencl.Buffer(self.ctx, mf.READ_WRITE, size=self.nb_channel*self.chunksize*self.dtype.itemsize)
        
        self.ring_sum_cl = pyopencl.Buffer(self.ctx, mf.READ_WRITE| mf.COPY_HOST_PTR, hostbuf=ring_sum)
        self.peak_mask_cl = pyopencl.Buffer(self.ctx, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=self.peak_mask)
        
        kernel = self.kernel%dict(chunksize=self.chunksize, nb_channel=self.nb_channel, n_span=self.n_span,
                    relative_threshold=self.relative_threshold, peak_sign={'+':1, '-':-1}[self.peak_sign])
        
        prg = pyopencl.Program(self.ctx, kernel)
        self.opencl_prg = prg.build(options='-cl-mad-enable')
        
        self.kern_detect_peaks = getattr(self.opencl_prg, 'detect_peaks')

    kernel = """
    #define chunksize %(chunksize)d
    #define n_span %(n_span)d
    #define nb_channel %(nb_channel)d
    #define relative_threshold %(relative_threshold)d
    #define peak_sign %(peak_sign)d
    
    __kernel void detect_peaks(__global  float *sigs,
                                                __global  float *ring_sum,
                                                __global  uchar *peaks){
    
        int pos = get_global_id(0);

        if (pos>=chunksize){
            return;
        }
        
        
        int idx;
        float v;

        // roll_ring_sum and wait for friends
        if (pos<(n_span*2)){
            ring_sum[pos] = ring_sum[pos+chunksize];
        }
        barrier(CLK_GLOBAL_MEM_FENCE);

        
        // sum all channels
        float sum=0;
        for (int chan=0; chan<nb_channel; chan++){
            idx = pos*nb_channel + chan;
            
            v = sigs[idx];
            
            //retify signals
            if(peak_sign==1){
                if (v<relative_threshold){v=0;}
            }
            else if(peak_sign==-1){
                if (v>-relative_threshold){v=0;}
            }
            
            sum = sum + v;
            
        }
        ring_sum[pos+2*n_span] = sum;
        
        barrier(CLK_GLOBAL_MEM_FENCE);
        
        
        // peaks span
        int pos2 = pos + n_span;
        
        uchar peak=0;
        
        if(peak_sign==1){
            if (ring_sum[pos2]>relative_threshold){
                peak=1;
                for (int i=1; i<=n_span; i++){
                    peak = peak && (ring_sum[pos2]>ring_sum[pos2-i]) && (ring_sum[pos2]>=ring_sum[pos2+i]);
                }
            }
        }
        else if(peak_sign==-1){
            if (ring_sum[pos2]<-relative_threshold){
                peak=1;
                for (int i=1; i<=n_span; i++){
                    peak = peak && (ring_sum[pos2]<ring_sum[pos2-i]) && (ring_sum[pos2]<=ring_sum[pos2+i]);
                }
            }
        }
        peaks[pos]=peak;
    
    }
    
    """


def get_mask_spatiotemporal_peaks(sigs, n_span, thresh, peak_sign, neighbours):
    sig_center = sigs[n_span:-n_span, :]

    if peak_sign == '+':
        mask_peaks = sig_center>thresh
        for chan in range(sigs.shape[1]):
            for neighbour in neighbours[chan, :]:
                if neighbour<0:
                    continue
                for i in range(n_span):
                    if chan != neighbour:
                        mask_peaks[:, chan] &= sig_center[:, chan] >= sig_center[:, neighbour]
                    mask_peaks[:, chan] &= sig_center[:, chan] > sigs[i:i+sig_center.shape[0], neighbour]
                    mask_peaks[:, chan] &= sig_center[:, chan]>=sigs[n_span+i+1:n_span+i+1+sig_center.shape[0], neighbour]
        
    elif peak_sign == '-':
        mask_peaks = sig_center<-thresh
        for chan in range(sigs.shape[1]):
            for neighbour in neighbours[chan, :]:
                if neighbour<0:
                    continue
                for i in range(n_span):
                    if chan != neighbour:
                        mask_peaks[:, chan] &= sig_center[:, chan] <= sig_center[:, neighbour]
                    mask_peaks[:, chan] &= sig_center[:, chan] < sigs[i:i+sig_center.shape[0], neighbour]
                    mask_peaks[:, chan] &= sig_center[:, chan]<=sigs[n_span+i+1:n_span+i+1+sig_center.shape[0], neighbour]
    
    return mask_peaks



class PeakDetectorGeometricalNumpy(BasePeakDetector):
    def process_buffer(self, sigs):
    
        if self.spatial_smooth_kernel is None:
            sigs = sigs
        else:
            sigs = np.dot(sigs, self.spatial_smooth_kernel)
        
        mask_peaks = self.get_mask_peaks_in_chunk(sigs)
        time_ind_peaks, chan_ind_peaks = np.nonzero(mask_peaks)
        time_ind_peaks += self.n_span
        peak_val_peaks = sigs[time_ind_peaks, chan_ind_peaks]
        
        return time_ind_peaks, chan_ind_peaks, peak_val_peaks
    
    def process_buffer_stream(self, pos, newbuf):
        self.fifo_sigs.new_chunk(newbuf, pos)
        sigs = self.fifo_sigs.get_data(pos-(newbuf.shape[0]+2*self.n_span), pos)
        
        time_ind_peaks, chan_ind_peaks, peak_val_peaks = self.process_buffer(sigs)

        if time_ind_peaks.size>0:
            time_ind_peaks += (pos - newbuf.shape[0] - 2 * self.n_span)
            return time_ind_peaks, chan_ind_peaks, peak_val_peaks

        return None, None, None
    
    def get_mask_peaks_in_chunk(self, fifo_residuals):
        
        # this is used by peeler engine geometry which handle externaly
        # a fifo residual
        mask_peaks = get_mask_spatiotemporal_peaks(fifo_residuals, self.n_span, self.relative_threshold, self.peak_sign, self.neighbours)
        return mask_peaks
    
    def change_params(self, adjacency_radius_um=200., **kargs):
        BasePeakDetector.change_params(self,  **kargs)
        
        self.adjacency_radius_um = adjacency_radius_um
        
        d = sklearn.metrics.pairwise.euclidean_distances(self.geometry)
        neighbour_mask = d<=self.adjacency_radius_um
        nb_neighbour_per_channel = np.sum(neighbour_mask, axis=0)
        nb_max_neighbour = np.max(nb_neighbour_per_channel)
        
        self.nb_max_neighbour = nb_max_neighbour # include itself
        self.neighbours = np.zeros((self.nb_channel, nb_max_neighbour), dtype='int32')
        
        self.neighbours[:] = -1
        for c in range(self.nb_channel):
            neighb, = np.nonzero(neighbour_mask[c, :])
            self.neighbours[c, :neighb.size] = neighb
        
        self.fifo_sigs = FifoBuffer((self.chunksize+2*self.n_span, self.nb_channel), self.dtype)
    
    def initialize_stream(self):
        self.fifo_sigs = FifoBuffer((self.chunksize+2*self.n_span, self.nb_channel), self.dtype)
    

class PeakDetectorGeometricalNumba(PeakDetectorGeometricalNumpy):
    def get_mask_peaks_in_chunk(self, fifo_residuals):
        mask_peaks = numba_get_mask_spatiotemporal_peaks(fifo_residuals, self.n_span, self.relative_threshold, self.peak_sign, self.neighbours)
        return mask_peaks
    


class PeakDetectorGeometricalOpenCL(PeakDetectorGeometricalNumpy, OpenCL_Helper):
    def process_buffer(self, sigs):
        
        if sigs.shape[0] <self.chunksize:
            sigs2 = np.zeros((self.chunksize, self.nb_channel), dtype=self.dtype)
            sigs2[:sigs.shape[0], :] = sigs
            sigs = sigs2
    
        if self.spatial_smooth_kernel is None:
            sigs = sigs
        else:
            sigs = np.dot(sigs, self.spatial_smooth_kernel)
        
        mask_peaks = self.get_mask_peaks_in_chunk(sigs)
        time_ind_peaks, chan_ind_peaks = np.nonzero(mask_peaks)
        time_ind_peaks += self.n_span
        peak_val_peaks = sigs[time_ind_peaks, chan_ind_peaks]
        
        return time_ind_peaks, chan_ind_peaks, peak_val_peaks
        
        
    def process_buffer_stream(self, pos, newbuf):
        if newbuf.shape[0] <self.chunksize:
            newbuf2 = np.zeros((self.chunksize, self.nb_channel), dtype=self.dtype)
            newbuf2[-newbuf.shape[0]:, :] = newbuf
            newbuf = newbuf2

        self.fifo_sigs.new_chunk(newbuf, pos)
        sigs = self.fifo_sigs.get_data(pos-(newbuf.shape[0]+2*self.n_span), pos)
        
        time_ind_peaks, chan_ind_peaks, peak_val_peaks = self.process_buffer(sigs)
        
        if time_ind_peaks.size>0:
            time_ind_peaks += (pos - newbuf.shape[0] - 2 * self.n_span)
            
            return time_ind_peaks, chan_ind_peaks, peak_val_peaks

        return None, None, None
    
    def get_mask_peaks_in_chunk(self, fifo_residuals):
        pyopencl.enqueue_copy(self.queue,  self.fifo_sigs_cl, fifo_residuals)
        #~ print(self.chunksize, self.max_wg_size)
        event = self.kern_get_mask_spatiotemporal_peaks(self.queue,  self.global_size, self.local_size,
                                self.fifo_sigs_cl, self.neighbours_cl, self.mask_peaks_cl)
        event.wait()
        pyopencl.enqueue_copy(self.queue,  self.mask_peaks, self.mask_peaks_cl)
        
        return self.mask_peaks
    
    def initialize_stream(self):
        self._make_gpu_buffer()
        
    def _make_gpu_buffer(self):
        # TODO fifo size should be : chunksize+2*self.n_span
        self.fifo_size = self.chunksize + 2*self.n_span
        self.fifo_sigs = FifoBuffer((self.fifo_size, self.nb_channel), self.dtype)
        self.mask_peaks = np.zeros((self.chunksize, self.nb_channel), dtype='uint8')  # bool
        
        #GPU buffers
        self.fifo_sigs_cl = pyopencl.Buffer(self.ctx, mf.READ_WRITE| mf.COPY_HOST_PTR, hostbuf=self.fifo_sigs.buffer)
        self.neighbours_cl = pyopencl.Buffer(self.ctx, mf.READ_WRITE| mf.COPY_HOST_PTR, hostbuf=self.neighbours)
        self.mask_peaks_cl = pyopencl.Buffer(self.ctx, mf.READ_WRITE| mf.COPY_HOST_PTR, hostbuf=self.mask_peaks)        
        
    
    def change_params(self, cl_platform_index=None, cl_device_index=None, **kargs):
        PeakDetectorGeometricalNumpy.change_params(self,  **kargs)
        
        OpenCL_Helper.initialize_opencl(self, cl_platform_index=cl_platform_index, cl_device_index=cl_device_index)
        #~ print(self.ctx)
        #~ print(self.chunksize)
        
        if self.chunksize > self.max_wg_size:
            n = int(np.ceil(self.chunksize / self.max_wg_size))
            self.global_size =  (self.max_wg_size * n, )
            self.local_size = (self.max_wg_size,)
        else:
            self.global_size = (self.chunksize, )
            self.local_size = (self.chunksize, )
        
        #~ print('self.global_size', self.global_size, 'self.chunksize', self.chunksize)
        
        
        self._make_gpu_buffer()
        
        # TODO fifo size should be : chunksize+2*self.n_span
        self.fifo_sigs = FifoBuffer((self.chunksize + 2*self.n_span, self.nb_channel), self.dtype)
        self.mask_peaks = np.zeros((self.chunksize, self.nb_channel), dtype='uint8')  # bool
        

        #GPU buffers
        self.fifo_sigs_cl = pyopencl.Buffer(self.ctx, mf.READ_WRITE| mf.COPY_HOST_PTR, hostbuf=self.fifo_sigs.buffer)
        self.neighbours_cl = pyopencl.Buffer(self.ctx, mf.READ_WRITE| mf.COPY_HOST_PTR, hostbuf=self.neighbours)
        self.mask_peaks_cl = pyopencl.Buffer(self.ctx, mf.READ_WRITE| mf.COPY_HOST_PTR, hostbuf=self.mask_peaks)
        
        kernel_ = opencl_kernel_geometrical_part1 + opencl_kernel_geometrical_part2
        
        kernel = kernel_ % dict(fifo_size=self.fifo_size, nb_channel=self.nb_channel, n_span=self.n_span,
                    relative_threshold=self.relative_threshold, peak_sign={'+':1, '-':-1}[self.peak_sign], nb_neighbour=self.nb_max_neighbour)
        
        prg = pyopencl.Program(self.ctx, kernel)

        self.opencl_prg = prg.build(options='-cl-mad-enable')
        
        self.kern_get_mask_spatiotemporal_peaks = getattr(self.opencl_prg, 'get_mask_spatiotemporal_peaks')

    
opencl_kernel_geometrical_part1 =  """
#define fifo_size %(fifo_size)d
#define n_span %(n_span)d
#define nb_channel %(nb_channel)d
#define relative_threshold %(relative_threshold)f
#define peak_sign %(peak_sign)d
#define nb_neighbour %(nb_neighbour)d
"""


opencl_kernel_geometrical_part2 =  """
__kernel void get_mask_spatiotemporal_peaks(__global  float *sigs,
                                            __global  int *neighbours,
                                            __global  uchar *mask_peaks){

    int pos = get_global_id(0);
    
    if (pos>=(fifo_size - (2 * n_span))){
        return;
    }
    

    float v;
    uchar peak;
    int chan_neigh;

    
    for (int chan=0; chan<nb_channel; chan++){
    
        v = sigs[(pos + n_span)*nb_channel + chan];

    
        if(peak_sign==1){
            if (v>relative_threshold){peak=1;}
            else if (v<=relative_threshold){peak=0;}
        }
        else if(peak_sign==-1){
            if (v<-relative_threshold){peak=1;}
            else if (v>=-relative_threshold){peak=0;}
        }
        
        if (peak == 1){
            for (int neighbour=0; neighbour<nb_neighbour; neighbour++){
                // when neighbour then chan==chan_neigh (itself)
                chan_neigh = neighbours[chan * nb_neighbour +neighbour];
                
                if (chan_neigh<0){continue;}
                
                if (chan != chan_neigh){
                    if(peak_sign==1){
                        peak = peak && (v>=sigs[(pos + n_span)*nb_channel + chan_neigh]);
                    }
                    else if(peak_sign==-1){
                        peak = peak && (v<=sigs[(pos + n_span)*nb_channel + chan_neigh]);
                    }
                }
                
                if (peak==0){break;}
                
                if(peak_sign==1){
                    for (int i=1; i<=n_span; i++){
                        peak = peak && (v>sigs[(pos + n_span - i)*nb_channel + chan_neigh]) && (v>=sigs[(pos + n_span + i)*nb_channel + chan_neigh]);
                        if (peak==0){break;}
                    }
                }
                else if(peak_sign==-1){
                    for (int i=1; i<=n_span; i++){
                        peak = peak && (v<sigs[(pos + n_span - i)*nb_channel + chan_neigh]) && (v<=sigs[(pos + n_span + i)*nb_channel + chan_neigh]);
                        if (peak==0){break;}
                    }
                }
            
            }
        }
        
        mask_peaks[pos * nb_channel + chan] = peak;
        
    }
}
"""


# TODO atomic_inc for spatiotemporal_opencl
# TODO rename engine propagate to GUI/examples/online

def get_peak_detector_class(method, engine):
    if engine == 'numba':
        assert HAVE_NUMBA, 'You must install numba'
    if engine == 'opencl':
        assert HAVE_PYOPENCL, 'You must install opencl'
    
    
    
    if method == 'global' and engine =='numba':
        print('WARNING : no peak detector global + numba use numpy instead')
        engine ='numpy'
    
    class_ = peakdetector_classes.get((method, engine))
    
    return class_
    


peakdetector_classes = { 
    ('global', 'numpy') : PeakDetectorGlobalNumpy, 
    ('global', 'opencl') : PeakDetectorGlobalOpenCL,
    ('geometrical', 'numpy') : PeakDetectorGeometricalNumpy,
    ('geometrical', 'numba') : PeakDetectorGeometricalNumba,
    ('geometrical', 'opencl'): PeakDetectorGeometricalOpenCL,
}





from .myqt import QT
import pyqtgraph as pg

import numpy as np

class TimeSeeker(QT.QWidget) :
    
    time_changed = QT.pyqtSignal(float)
    
    def __init__(self, parent = None, show_slider = True, show_spinbox = True) :
        QT.QWidget.__init__(self, parent)
        
        self.layout = QT.QHBoxLayout()
        self.setLayout(self.layout)
        
        if show_slider:
            self.slider = QT.QSlider(orientation=QT.Qt.Horizontal, minimum=0, maximum=999)
            self.layout.addWidget(self.slider)
            self.slider.valueChanged.connect(self.slider_changed)
        else:
            self.slider = None
            
        if show_spinbox:
            self.spinbox = pg.SpinBox(decimals = 4, bounds=[-np.inf, np.inf], suffix = 's', siPrefix = True, 
                            step = 0.1, dec = True, minStep = 0.001)
            self.layout.addWidget(self.spinbox)
            self.spinbox.valueChanged.connect(self.spinbox_changed)
        else:
            self.spinbox = None

        self.t = 0 #  s
        self.set_start_stop(0., 10.)

    def set_start_stop(self, t_start, t_stop, seek = True):
        if np.isnan(t_start) or np.isnan(t_stop): return
        assert t_stop>t_start
        self.t_start = t_start
        self.t_stop = t_stop
        
        if seek:
            self.seek(self.t_start)
        
        if self.spinbox is not None:
            self.spinbox.setMinimum(t_start)
            self.spinbox.setMaximum(t_stop)

    def slider_changed(self, pos):
        t = pos/1000.*(self.t_stop - self.t_start)+self.t_start
        self.seek(t, set_slider = False)
    
    def spinbox_changed(self, val):
        self.seek(val, set_spinbox = False)
        
    def seek(self, t, set_slider = True, set_spinbox = True, emit = True):
        self.t = t
        
        if self.slider is not None and set_slider:
            self.slider.valueChanged.disconnect(self.slider_changed)
            pos = int((self.t - self.t_start)/(self.t_stop - self.t_start)*1000.)
            self.slider.setValue(pos)
            self.slider.valueChanged.connect(self.slider_changed)
        
        if self.spinbox is not None and set_spinbox:
            self.spinbox.valueChanged.disconnect(self.spinbox_changed)
            self.spinbox.setValue(t)
            self.spinbox.valueChanged.connect(self.spinbox_changed)
        
        if emit:
            self.time_changed.emit(float(self.t))

def get_dict_from_group_param(param, cascade = False):
    assert param.type() == 'group'
    d = {}
    for p in param.children():
        if p.type() == 'group':
            if cascade:
                d[p.name()] = get_dict_from_group_param(p, cascade = True)
            continue
        else:
            d[p.name()] = p.value()
    return d

def set_group_param_from_dict(param, d, cascade=False):
    assert param.type() == 'group'
    for p in param.children():
        k = p.name()
        if p.type() == 'group':
            if cascade:
                set_group_param_from_dict(p, d[k], cascade=True)
            continue
        else:
            if k in d:
                if d[k] is None and p.type() == 'float':
                        param[k] = np.nan
                else:
                    param[k] = d[k]
    return d
    

class ParamDialog(QT.QDialog):
    def __init__(self,   params, title = '', parent = None):
        QT.QDialog.__init__(self, parent = parent)
        
        self.setWindowTitle(title)
        self.setModal(True)
        
        self.params = pg.parametertree.Parameter.create( name=title, type='group', children = params)
        
        layout = QT.QVBoxLayout()
        self.setLayout(layout)

        self.tree_params = pg.parametertree.ParameterTree(parent  = self)
        self.tree_params.header().hide()
        self.tree_params.setParameters(self.params, showTop=True)
        #self.tree_params.setWindowFlags(QT.Qt.Window)
        layout.addWidget(self.tree_params)

        but = QT.QPushButton('OK')
        layout.addWidget(but)
        but.clicked.connect(self.accept)
        
        but.setFocus()

    def get(self):
        return get_dict_from_group_param(self.params, cascade=True)
    
    def set(self, d):
        set_group_param_from_dict(self.params, d, cascade=True)


class MethodDialog(QT.QDialog):
    def __init__(self,   params_by_method, title = '', parent = None, selected_method=None):
        QT.QDialog.__init__(self, parent = parent)
        
        self.setWindowTitle(title)
        self.setModal(True)

        layout = QT.QVBoxLayout()
        self.setLayout(layout)
        
        methods = list(params_by_method.keys())
        if selected_method is not None:
            assert selected_method in methods
        self.methods = methods
        
        params = [{'name' : 'method', 'type' : 'list', 'values' : methods}]
        self.param_method = pg.parametertree.Parameter.create( name=title, type='group', children = params)
        self.tree_params = pg.parametertree.ParameterTree(parent  = self)
        self.tree_params.header().hide()
        self.tree_params.setParameters(self.param_method, showTop=True)
        #self.tree_params.setWindowFlags(QT.Qt.Window)
        layout.addWidget(self.tree_params, 1)
        
        
        self.all_params = {}
        self.all_tree_params = {}
        for method in methods:
            params = params_by_method[method]
            self.all_params[method] =  pg.parametertree.Parameter.create(name='params', type='group', children=params)
            tree = pg.parametertree.ParameterTree(parent=self)
            tree.header().hide()
            tree.setParameters(self.all_params[method], showTop=True)
            layout.addWidget(tree, 5)
            tree.hide()
            self.all_tree_params[method] = tree

        but = QT.QPushButton('OK')
        layout.addWidget(but)
        but.clicked.connect(self.accept)
        but.setFocus()
        
        if selected_method is None:
            selected_method = methods[0]
        
        self.param_method.sigTreeStateChanged.connect(self.on_method_change)
        self.param_method['method'] = selected_method
        self.all_tree_params[selected_method].show()
    
    def on_method_change(self):
        #~ print('on_method_change')
        for tree in self.all_tree_params.values():
            tree.hide()
        
        method =  self.param_method['method']
        #~ print(method)
        self.all_tree_params[method].show()
    
    def set_method(self, method, d):
        self.param_method['method'] = method
        set_group_param_from_dict(self.all_params[method], d, cascade=True)
    
    
    def get(self):
        method = self.param_method['method']
        d = get_dict_from_group_param(self.all_params[method], cascade=True)
        return method, d

def open_dialog_methods(params_by_method, parent, title='Which method ?', selected_method=None):
        
        dia = MethodDialog(params_by_method, parent=parent, title=title, selected_method=selected_method)
        
        if dia.exec_():
            method = dia.param_method['method']
            kargs = get_dict_from_group_param(dia.all_params[method], cascade=True)
            return method, kargs
        else:
            return None, None
        


if __name__=='__main__':
    app = pg.mkQApp()
    #~ timeseeker =TimeSeeker()
    #~ timeseeker.show()
    #~ app.exec_()
    params = [{'name' : 'a', 'value' : 1., 'type' : 'float'}]
    dialog = ParamDialog(params, title = 'yep')
    dialog.exec_()
    print(dialog.get())
    
    





from pprint import pprint
import os


import numpy as np
import pyqtgraph as pg


from tridesclous.gui.tools import get_dict_from_group_param
from tridesclous.gui.gui_params import preprocessor_params, peak_detector_params, clean_peaks_params



preprocessor_params_default = get_dict_from_group_param(
                pg.parametertree.Parameter.create(name='',
                    type='group', children=preprocessor_params))


peak_detector_params_default = get_dict_from_group_param(
                pg.parametertree.Parameter.create(name='',
                    type='group', children=peak_detector_params))

clean_peaks_params_default = get_dict_from_group_param(
                pg.parametertree.Parameter.create(name='',
                    type='group', children=clean_peaks_params))


    

def make_empty_catalogue(chan_grp=0,
                channel_indexes=[],
                n_left=-20,
                n_right=40,
                n_left_long=-40,
                n_right_long=80,
                internal_dtype='float32',
                
                preprocessor_params={},
                peak_detector_params={},
                clean_peaks_params={},
                
                signals_medians = None,
                signals_mads = None,
                
                ):
    catalogue = {}
    
    catalogue['chan_grp'] = chan_grp
    catalogue['n_left'] = n_left
    catalogue['n_right'] = n_right
    catalogue['n_left_long'] = n_left_long
    catalogue['n_right_long'] = n_right_long
    catalogue['peak_width'] = catalogue['n_right'] - catalogue['n_left']
    
    catalogue['cluster_labels'] = np.array([], dtype='int64')
    catalogue['clusters'] = np.array([], dtype='int64')
    
    catalogue['mode'] = 'sparse'
    
    full_width = n_right - n_left
    nchan = len(channel_indexes)
    
    centers0 = np.zeros((0, full_width - 4, nchan), dtype=internal_dtype)
    centers1 = np.zeros_like(centers0)
    centers2 = np.zeros_like(centers0)
    catalogue['centers0'] = centers0
    catalogue['centers1'] = centers1
    catalogue['centers2'] = centers2
    
    subsample = np.arange(1.5, full_width-2.5, 1/20.)
    catalogue['subsample_ratio'] = 20
    interp_centers0 = np.zeros((0, subsample.size, nchan), dtype=internal_dtype)
    catalogue['interp_centers0'] = interp_centers0
    
    catalogue['label_to_index'] = {}

    #find max  channel for each cluster for peak alignement
    catalogue['extremum_channel'] = np.zeros_like(catalogue['cluster_labels'])
    
    
    preprocessor_params_ = dict(preprocessor_params_default)
    preprocessor_params_.update(preprocessor_params)
    #~ preprocessor_params_.pop('chunksize')
    


    peak_detector_params_ = dict(peak_detector_params_default)
    peak_detector_params_.update(peak_detector_params)

    clean_peaks_params_ = dict(clean_peaks_params_default)
    clean_peaks_params_.update(clean_peaks_params)
    
    if signals_medians is None:
        signals_medians = signals_medians = np.zeros(nchan, dtype=internal_dtype)
    if signals_mads is None:
        signals_mads = signals_mads = np.ones(nchan, dtype=internal_dtype)
    
    catalogue['inter_sample_oversampling'] = False
    
    catalogue['sparse_mask_level1'] = np.ones((0, nchan), dtype='bool')
    catalogue['sparse_mask_level2'] = np.ones((0, nchan), dtype='bool')
    catalogue['sparse_mask_level3'] = np.ones((0, nchan), dtype='bool')
    
    catalogue['distance_limit'] = np.zeros((0,), dtype='float32')
    
    
    #params
    catalogue['signal_preprocessor_params'] = preprocessor_params_
    catalogue['peak_detector_params'] = peak_detector_params_
    catalogue['clean_peaks_params'] = clean_peaks_params_
    catalogue['signals_medians'] = signals_medians
    catalogue['signals_mads'] = signals_mads
    
    catalogue['empty_catalogue'] = True # a key to detect real/fake catalogue

    catalogue['projections'] = np.zeros((0,), dtype='float32')
    catalogue['boundaries'] = np.zeros((0,), dtype='float32')
    
    
    return catalogue


def lighter_catalogue(catalogue):
    """
    The trace viewer need the catalogue for some fiew kargs.
    Since with pyacq rpc system all kargs are serializer since function
    make the catalogue lightened with only needed keys
    """
    lightened_catalogue = {}
    keys = ['clusters', 'cluster_labels', 'extremum_channel']
    for k in keys:
        lightened_catalogue[k] = catalogue[k]
    
    return lightened_catalogue




import os
import numpy as np
import re
from collections import OrderedDict


data_source_classes = OrderedDict()

possible_modes = ['one-file', 'multi-file', 'one-dir', 'multi-dir', 'other']


import neo



class DataSourceBase:
    gui_params = None
    def __init__(self):
        # important total_channel != nb_channel because nb_channel=len(channels)
        self.total_channel = None 
        self.sample_rate = None
        self.nb_segment = None
        self.dtype = None
        self.bit_to_microVolt = None
    
    def get_segment_shape(self, seg_num):
        raise NotImplementedError
    
    def get_channel_names(self):
        raise NotImplementedError
    
    def get_signals_chunk(self, seg_num=0, i_start=None, i_stop=None):
        raise NotImplementedError
    

    
class InMemoryDataSource(DataSourceBase):
    """
    DataSource in memory numpy array.
    This is for debugging  or fast testing.
    """
    mode = 'other'
    def __init__(self, nparrays=[], sample_rate=None):
        DataSourceBase.__init__(self)
        
        self.nparrays = nparrays
        self.nb_segment = len(self.nparrays)
        self.total_channel = self.nparrays[0].shape[1]
        self.sample_rate = sample_rate
        self.dtype = self.nparrays[0].dtype
    
    def get_segment_shape(self, seg_num):
        full_shape = self.nparrays[seg_num].shape
        return full_shape
    
    def get_signals_chunk(self, seg_num=0, i_start=None, i_stop=None):
            data = self.nparrays[seg_num][i_start:i_stop, :]
            return data
            
    def get_channel_names(self):
        return ['ch{}'.format(i) for i in range(self.total_channel)]

data_source_classes['InMemory'] = InMemoryDataSource





class RawDataSource(DataSourceBase):
    """
    DataSource from raw binary file. Easy case.
    """
    mode = 'multi-file'
    gui_params = [
        {'name': 'dtype', 'type': 'list', 'values':['int16', 'uint16', 'float32', 'float64']},
        {'name': 'total_channel', 'type': 'int', 'value':1},
        {'name': 'sample_rate', 'type': 'float', 'value':10000., 'step': 1000., 'suffix': 'Hz', 'siPrefix': True},
        {'name': 'offset', 'type': 'int', 'value':0},
        {'name': 'bit_to_microVolt', 'type': 'float', 'value':0.5 },        
    ]
    
    def __init__(self, filenames=[], dtype='int16', total_channel=0,
                        sample_rate=0., offset=0, bit_to_microVolt=None, channel_names=None):
        DataSourceBase.__init__(self)
        
        self.filenames = filenames
        if isinstance(self.filenames, str):
            self.filenames = [self.filenames]
        assert all([os.path.exists(f) for f in self.filenames]), 'files does not exist'
        self.nb_segment = len(self.filenames)

        self.total_channel = total_channel
        self.sample_rate = sample_rate
        self.dtype = np.dtype(dtype)
        
        if bit_to_microVolt == 0.:
            bit_to_microVolt = None
        self.bit_to_microVolt = bit_to_microVolt
        
        if channel_names is None:
            channel_names = ['ch{}'.format(i) for i in range(self.total_channel)]
        self.channel_names = channel_names
        

        self.array_sources = []
        for filename in self.filenames:
            data = np.memmap(filename, dtype=self.dtype, mode='r', offset=offset)
            #~ data = data[:-(data.size%self.total_channel)]
            data = data.reshape(-1, self.total_channel)
            self.array_sources.append(data)
    
    def get_segment_shape(self, seg_num):
        full_shape = self.array_sources[seg_num].shape
        return full_shape
    
    def get_signals_chunk(self, seg_num=0, i_start=None, i_stop=None):
            data = self.array_sources[seg_num][i_start:i_stop, :]
            return data

    def get_channel_names(self):
        return self.channel_names

data_source_classes['RawData'] = RawDataSource






import neo.rawio

io_gui_params = {
    'RawBinarySignal':[
                {'name': 'dtype', 'type': 'list', 'values':['int16', 'uint16', 'float32', 'float64']},
                {'name': 'nb_channel', 'type': 'int', 'value':1},
                {'name': 'sampling_rate', 'type': 'float', 'value':10000., 'step': 1000., 'suffix': 'Hz', 'siPrefix': True},
                {'name': 'bytesoffset', 'type': 'int', 'value':0},
    ],
}


# hook for some neo.rawio that have problem with TDC (multi sampling rate or default params)
neo_rawio_hooks = {}

class Intan(neo.rawio.IntanRawIO):
    def _parse_header(self):
        neo.rawio.IntanRawIO._parse_header(self)
        sig_channels = self.header['signal_channels']
        sig_channels = sig_channels[sig_channels['group_id']==0]
        self.header['signal_channels'] = sig_channels

    def _get_signal_size(self, block_index, seg_index, channel_indexes):
        if channel_indexes is None:
            channel_indexes = slice(None)
        assert np.unique(self.header['signal_channels'][channel_indexes]['group_id']).size == 1
        channel_names = self.header['signal_channels'][channel_indexes]['name']
        chan_name = channel_names[0]
        size = self._raw_data[chan_name].size
        return size

neo_rawio_hooks['Intan'] = Intan



class NeoRawIOAggregator(DataSourceBase):
    """
    wrappe and agregate several neo.rawio in the class.
    """
    gui_params = None
    rawio_class = None
    def __init__(self, **kargs):
        DataSourceBase.__init__(self)
        
        self.rawios = []
        if  'filenames' in kargs:
            filenames= kargs.pop('filenames') 
            self.rawios = [self.rawio_class(filename=f, **kargs) for f in filenames]
        elif 'dirnames' in kargs:
            dirnames= kargs.pop('dirnames') 
            self.rawios = [self.rawio_class(dirname=d, **kargs) for d in dirnames]
        else:
            raise(ValueError('Must have filenames or dirnames'))
            
        
        self.sample_rate = None
        self.total_channel = None
        self.sig_channels = None
        nb_seg = 0
        self.segments = {}
        for rawio in self.rawios:
            rawio.parse_header()
            assert not rawio._several_channel_groups, 'several sample rate for signals'
            assert rawio.block_count() ==1, 'Multi block RawIO not implemented'
            for s in range(rawio.segment_count(0)):
                #nb_seg = absolut seg index and s= local seg index
                self.segments[nb_seg] = (rawio, s)
                nb_seg += 1
            
            if self.sample_rate is None:
                self.sample_rate = rawio.get_signal_sampling_rate()
            else:
                assert self.sample_rate == rawio.get_signal_sampling_rate(), 'bad joke different sample rate!!'
            
            sig_channels = rawio.header['signal_channels']
            if self.sig_channels is None:
                self.sig_channels = sig_channels
                self.total_channel = len(sig_channels)
            else:
                assert np.all(sig_channels==self.sig_channels), 'bad joke different channels!'
            
        self.nb_segment = len(self.segments)
        
        self.dtype = np.dtype(self.sig_channels['dtype'][0])
        units = sig_channels['units'][0]
        #~ assert 'V' in units, 'Units are not V, mV or uV'
        if units =='V':
            self.bit_to_microVolt = self.sig_channels['gain'][0]*1e-6
        elif units =='mV':
            self.bit_to_microVolt = self.sig_channels['gain'][0]*1e-3
        elif units =='uV':
            self.bit_to_microVolt = self.sig_channels['gain'][0]
        else:
            self.bit_to_microVolt = None
        
    def get_segment_shape(self, seg_num):
        rawio, s = self.segments[seg_num]
        l = rawio.get_signal_size(0, s)
        return l, self.total_channel
    
    def get_channel_names(self):
        return self.sig_channels['name'].tolist()
    
    def get_signals_chunk(self, seg_num=0, i_start=None, i_stop=None):
        rawio, s = self.segments[seg_num]
        return rawio.get_analogsignal_chunk(block_index=0, seg_index=s, 
                        i_start=i_start, i_stop=i_stop)

#Construct the list with taking local class with hooks dict
rawiolist = []
for rawio_class in neo.rawio.rawiolist:
    name = rawio_class.__name__.replace('RawIO', '')
    if name in neo_rawio_hooks:
        rawio_class = neo_rawio_hooks[name]
    rawiolist.append(rawio_class)

if neo.rawio.RawBinarySignalRawIO in rawiolist:
    # to avoid bug in readthe doc with moc
    RawBinarySignalRawIO = rawiolist.pop(rawiolist.index(neo.rawio.RawBinarySignalRawIO))
#~ rawiolist.insert(0, RawBinarySignalRawIO)

for rawio_class in rawiolist:
    name = rawio_class.__name__.replace('RawIO', '')
    class_name = name+'DataSource'
    datasource_class = type(class_name,(NeoRawIOAggregator,), { })
    datasource_class.rawio_class = rawio_class
    if rawio_class.rawmode in ('multi-file', 'one-file'):
        #multi file in neo have another meaning
        datasource_class.mode = 'multi-file'
    elif rawio_class.rawmode in ('one-dir', ):
        datasource_class.mode = 'multi-dir'
    else:
        continue
    
    #gui stuffs
    if name in io_gui_params:
        datasource_class.gui_params = io_gui_params[name]
        
    data_source_classes[name] = datasource_class
    #~ print(datasource_class, datasource_class.mode )


    
#TODO implement KWIK and OpenEphys
#https://open-ephys.atlassian.net/wiki/display/OEW/Data+format
# https://github.com/open-ephys/analysis-tools/tree/master/Python3





from .myqt import QT
import pyqtgraph as pg

import numpy as np
import pandas as pd

from .base import WidgetBase
from ..tools import compute_cross_correlograms


class MyViewBox(pg.ViewBox):
    doubleclicked = QT.pyqtSignal()
    def mouseDoubleClickEvent(self, ev):
        self.doubleclicked.emit()
        ev.accept()
    def raiseContextMenu(self, ev):
        #for some reasons enableMenu=False is not taken (bug ????)
        pass



class CrossCorrelogramViewer(WidgetBase):
    _params = [
                      {'name': 'window_size_ms', 'type': 'float', 'value' : 100. },
                      {'name': 'bin_size_ms', 'type': 'float', 'value' : 1.0 },
                      {'name': 'symmetrize', 'type': 'bool', 'value' : True },
                      {'name': 'display_axis', 'type': 'bool', 'value' : True },
                      {'name': 'max_visible', 'type': 'int', 'value' : 8 },
                      {'name': 'check_sorted', 'type': 'bool', 'value' : False },
        ]
    def __init__(self, controller=None, parent=None):
        WidgetBase.__init__(self, parent=parent, controller=controller)
        
        self.layout = QT.QVBoxLayout()
        self.setLayout(self.layout)
        
        h = QT.QHBoxLayout()
        self.layout.addLayout(h)

        but = QT.QPushButton('settings')
        but.clicked.connect(self.open_settings)
        h.addWidget(but)

        but = QT.QPushButton('compute')
        but.clicked.connect(self.compute_ccg)
        h.addWidget(but)
        
        self.grid = pg.GraphicsLayoutWidget()
        self.layout.addWidget(self.grid)
        
        self.ccg = None


    def on_params_changed(self):
        self.ccg = None
        self.refresh()
    
    def compute_ccg(self):
        cluster_labels = self.controller.positive_cluster_labels
        spikes = self.controller.spikes
        try:
            # this prevent some bug like this one https://github.com/tridesclous/tridesclous/issues/69
            self.ccg, self.bins = compute_cross_correlograms(spikes['index'], spikes['cluster_label'],
                        spikes['segment'], cluster_labels, self.controller.dataio.sample_rate,
                        window_size=self.params['window_size_ms']/1000.,
                        bin_size = self.params['bin_size_ms']/1000.,
                        symmetrize=self.params['symmetrize'],
                        check_sorted=self.params['check_sorted'],
                        )
        except:
            self.ccg, self.bins = None, None
        self.refresh()

    def refresh(self):
        self.grid.clear()
        
        if self.ccg is None:
            return
        
        visibles = [ ]
        for k in self.controller.positive_cluster_labels:
            if self.controller.cluster_visible[k]:
                visibles.append(k)
        
        visibles = visibles[:self.params['max_visible']]
        
        n = len(visibles)
        
        bins = self.bins * 1000. #to ms
        
        labels = self.controller.positive_cluster_labels.tolist()
        
        for r in range(n):
            for c in range(r, n):
                
                i = labels.index(visibles[r])
                j = labels.index(visibles[c])
                
                count = self.ccg[i, j, :]
                
                plot = pg.PlotItem()
                if not self.params['display_axis']:
                    plot.hideAxis('bottom')
                    plot.hideAxis('left')
                
                if r==c:
                    k = visibles[r]
                    color = self.controller.qcolors[k]
                else:
                    color = (120,120,120,120)
                
                curve = pg.PlotCurveItem(bins, count, stepMode=True, fillLevel=0, brush=color, pen=color)
                plot.addItem(curve)
                self.grid.addItem(plot, row=r, col=c)


#~ plt1.plot(x, y, stepMode=True, fillLevel=0, brush=(0,0,255,150))




import numpy as np

from .myqt import QT
import pyqtgraph as pg

from ..catalogueconstructor import _default_n_spike_for_centroid
from ..export import export_catalogue_spikes

from .cataloguecontroller import CatalogueController
from .traceviewer import CatalogueTraceViewer
from .peaklists import PeakList, ClusterPeakList
from .ndscatter import NDScatter
from .waveformviewer import WaveformViewer
from .similarity import SpikeSimilarityView, ClusterSimilarityView, ClusterRatioSimilarityView
from .pairlist import PairList
from .silhouette import Silhouette
from .waveformhistviewer import WaveformHistViewer
from .featuretimeviewer import FeatureTimeViewer

from .tools import ParamDialog, open_dialog_methods

from . import gui_params

from . import icons


import itertools
import datetime
import time
import webbrowser
from pprint import pprint

class CatalogueWindow(QT.QMainWindow):
    new_catalogue = QT.pyqtSignal(int)
    def __init__(self, catalogueconstructor):
        QT.QMainWindow.__init__(self)
        
        self.setWindowIcon(QT.QIcon(':/main_icon.png'))
        
        self.catalogueconstructor = catalogueconstructor
        self.controller = CatalogueController(catalogueconstructor=catalogueconstructor)

        self.traceviewer = CatalogueTraceViewer(controller=self.controller)
        self.peaklist = PeakList(controller=self.controller)
        self.clusterlist = ClusterPeakList(controller=self.controller)
        self.ndscatter = NDScatter(controller=self.controller)
        self.waveformviewer = WaveformViewer(controller=self.controller)
        #~ self.spikesimilarityview = SpikeSimilarityView(controller=self.controller)
        self.clustersimilarityview = ClusterSimilarityView(controller=self.controller)
        self.clusterratiosimilarityview = ClusterRatioSimilarityView(controller=self.controller)
        self.pairlist = PairList(controller=self.controller)
        self.silhouette = Silhouette(controller=self.controller)
        self.waveformhistviewer = WaveformHistViewer(controller=self.controller)
        self.featuretimeviewer = FeatureTimeViewer(controller=self.controller)
        
        docks = {}

        docks['waveformviewer'] = QT.QDockWidget('waveformviewer',self)
        docks['waveformviewer'].setWidget(self.waveformviewer)
        #self.tabifyDockWidget(docks['ndscatter'], docks['waveformviewer'])
        self.addDockWidget(QT.Qt.RightDockWidgetArea, docks['waveformviewer'])
        

        docks['waveformhistviewer'] = QT.QDockWidget('waveformhistviewer',self)
        docks['waveformhistviewer'].setWidget(self.waveformhistviewer)
        self.tabifyDockWidget(docks['waveformviewer'], docks['waveformhistviewer'])

        docks['featuretimeviewer'] = QT.QDockWidget('featuretimeviewer',self)
        docks['featuretimeviewer'].setWidget(self.featuretimeviewer)
        self.tabifyDockWidget(docks['waveformhistviewer'], docks['featuretimeviewer'])
        
        
        docks['traceviewer'] = QT.QDockWidget('traceviewer',self)
        docks['traceviewer'].setWidget(self.traceviewer)
        #self.addDockWidget(QT.Qt.RightDockWidgetArea, docks['traceviewer'])
        self.tabifyDockWidget(docks['waveformviewer'], docks['traceviewer'])
        
        docks['peaklist'] = QT.QDockWidget('peaklist',self)
        docks['peaklist'].setWidget(self.peaklist)
        self.addDockWidget(QT.Qt.LeftDockWidgetArea, docks['peaklist'])
        
        docks['pairlist'] = QT.QDockWidget('pairlist',self)
        docks['pairlist'].setWidget(self.pairlist)
        self.splitDockWidget(docks['peaklist'], docks['pairlist'], QT.Qt.Horizontal)
        
        docks['clusterlist'] = QT.QDockWidget('clusterlist',self)
        docks['clusterlist'].setWidget(self.clusterlist)
        self.tabifyDockWidget(docks['pairlist'], docks['clusterlist'])
        
        #on bottom left
        #~ docks['spikesimilarityview'] = QT.QDockWidget('spikesimilarityview',self)
        #~ docks['spikesimilarityview'].setWidget(self.spikesimilarityview)
        #~ self.addDockWidget(QT.Qt.LeftDockWidgetArea, docks['spikesimilarityview'])

        docks['clustersimilarityview'] = QT.QDockWidget('clustersimilarityview',self)
        docks['clustersimilarityview'].setWidget(self.clustersimilarityview)
        #~ self.tabifyDockWidget(docks['spikesimilarityview'], docks['clustersimilarityview'])
        self.addDockWidget(QT.Qt.LeftDockWidgetArea, docks['clustersimilarityview'])

        docks['clusterratiosimilarityview'] = QT.QDockWidget('clusterratiosimilarityview',self)
        docks['clusterratiosimilarityview'].setWidget(self.clusterratiosimilarityview)
        #~ self.tabifyDockWidget(docks['spikesimilarityview'], docks['clusterratiosimilarityview'])
        self.tabifyDockWidget(docks['clustersimilarityview'], docks['clusterratiosimilarityview'])
        

        docks['silhouette'] = QT.QDockWidget('silhouette',self)
        docks['silhouette'].setWidget(self.silhouette)
        #~ self.tabifyDockWidget(docks['spikesimilarityview'], docks['silhouette'])
        self.tabifyDockWidget(docks['clustersimilarityview'], docks['silhouette'])
        
        
        docks['ndscatter'] = QT.QDockWidget('ndscatter',self)
        docks['ndscatter'].setWidget(self.ndscatter)
        #~ self.tabifyDockWidget(docks['spikesimilarityview'], docks['ndscatter'])
        self.tabifyDockWidget(docks['clustersimilarityview'], docks['ndscatter'])
        
        self.create_actions()
        self.create_toolbar()
        
        
        
    def create_actions(self):

        self.act_make_catalogue = QT.QAction('Make catalogue for peeler', self,checkable = False, icon=QT.QIcon(":/document-save.svg"))
        self.act_make_catalogue.triggered.connect(self.make_catalogue_for_peeler)

        self.act_savepoint = QT.QAction('Savepoint', self,checkable = False, icon=QT.QIcon(":/document-save.svg"))
        self.act_savepoint.triggered.connect(self.create_savepoint)
        
        #~ self.act_refresh = QT.QAction('Refresh', self,checkable = False, icon=QT.QIcon.fromTheme("view-refresh"))
        self.act_refresh = QT.QAction('Refresh', self,checkable = False, icon=QT.QIcon(":/view-refresh.svg"))
        self.act_refresh.triggered.connect(self.refresh_with_reload)
        
        self.act_redetect_peak = QT.QAction('New peaks', self,checkable = False, icon=QT.QIcon(":/configure-shortcuts.svg"))
        self.act_redetect_peak.triggered.connect(self.redetect_peak)

        self.act_new_waveform_sample = QT.QAction('New waveform sample', self,checkable = False, icon=QT.QIcon(":/configure-shortcuts.svg"))
        self.act_new_waveform_sample.triggered.connect(self.new_waveform_sample)
        
        self.act_new_features = QT.QAction('New features', self,checkable = False, icon=QT.QIcon(":/configure-shortcuts.svg"))
        self.act_new_features.triggered.connect(self.new_features)

        self.act_new_cluster = QT.QAction('New cluster', self,checkable = False, icon=QT.QIcon(":/configure-shortcuts.svg"))
        self.act_new_cluster.triggered.connect(self.new_cluster)

        self.act_compute_metrics = QT.QAction('Compute metrics', self,checkable = False, icon=QT.QIcon(":/configure-shortcuts.svg"))
        self.act_compute_metrics.triggered.connect(self.compute_metrics)


        
        self.help_act = QT.QAction('Help', self,checkable = False, icon=QT.QIcon(":main_icon.png"))
        self.help_act.triggered.connect(self.open_webbrowser_help)
        


    def create_toolbar(self):
        self.toolbar = QT.QToolBar('Tools')
        self.toolbar.setToolButtonStyle(QT.Qt.ToolButtonTextUnderIcon)
        self.addToolBar(QT.Qt.RightToolBarArea, self.toolbar)
        self.toolbar.setIconSize(QT.QSize(60, 40))
        
        
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.act_refresh)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.act_redetect_peak)
        self.toolbar.addAction(self.act_new_waveform_sample)
        self.toolbar.addAction(self.act_new_features)
        self.toolbar.addAction(self.act_new_cluster)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.act_make_catalogue)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.act_compute_metrics)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.help_act)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.act_savepoint)
        
        # export peaks
        self.toolbar.addSeparator()
        menu = QT.QMenu()
        toolButton = QT.QToolButton()
        toolButton.setMenu(menu)
        toolButton.setPopupMode(QT.QToolButton.InstantPopup)
        self.toolbar.addWidget(toolButton)
        do_export_peaks = QT.QAction('Export peaks', self)
        do_export_peaks.triggered.connect(self.export_peaks)
        menu.addAction(do_export_peaks)    

    def warn(self, title, text):
        mb = QT.QMessageBox.warning(self, title,text, QT.QMessageBox.Ok ,  QT.QMessageBox.NoButton)
    
    def open_webbrowser_help(self):
        url = "http://tridesclous.readthedocs.io/en/latest/catalogue_window.html"
        webbrowser.open(url, new=2)

    def make_catalogue_for_peeler(self):
        dia = ParamDialog(gui_params.make_catalogue_params)
        dia.resize(450, 500)
        if dia.exec_():
            d = dia.get()
            pprint(d)
            self.catalogueconstructor.make_catalogue_for_peeler(**d)
            self.new_catalogue.emit(self.catalogueconstructor.chan_grp)
    
    def create_savepoint(self):
        try:
            copy_path = self.catalogueconstructor.create_savepoint()
        except:
            copy_path = None
        
        if copy_path is None:
            txt = 'Savepoint FAIL!!!'
        else:
            txt = 'Savepoint done here {}'.format(copy_path)
        self.warn('savepoint', txt)
        
    
    def refresh_with_reload(self):
        self.controller.reload_data()
        self.refresh()
    
    def refresh(self):
        self.controller.check_plot_attributes()
        for w in self.controller.views:
            #TODO refresh only visible but need catch on visibility changed
            #~ print(w)
            #~ t1 = time.perf_counter()
            w.refresh()
            #~ t2 = time.perf_counter()
            #~ print('refresh',w,  t2-t1)
    
    def redetect_peak(self):
        dia = ParamDialog(gui_params.peak_detector_params)
        dia.resize(450, 500)
        d = self.catalogueconstructor.info['peak_detector']
        dia.set(d)
        
        if dia.exec_():
            d = dia.get()
            self.catalogueconstructor.re_detect_peak(**d)
            self.controller.init_plot_attributes()
        self.refresh()
    
    #~ def new_waveforms(self):
        #~ dia = ParamDialog(gui_params.waveforms_params)
        #~ dia.resize(450, 500)
        #~ if dia.exec_():
            #~ d = dia.get()
            #~ self.catalogueconstructor.extract_some_waveforms(**d)
            
            #~ self.refresh()

    #~ def clean_waveforms(self):
        #~ dia = ParamDialog(gui_params.clean_waveforms_params)
        #~ dia.resize(450, 500)
        #~ if dia.exec_():
            #~ d = dia.get()
            #~ self.catalogueconstructor.clean_waveforms(**d)
            #~ self.refresh()

    #~ def new_noise_snippet(self):
        #~ dia = ParamDialog(gui_params.noise_snippet_params)
        #~ dia.resize(450, 500)
        #~ if dia.exec_():
            #~ d = dia.get()
            #~ self.catalogueconstructor.extract_some_noise(**d)
        #~ self.refresh()
    
    def new_waveform_sample(self):
        params_ = [
            {'name':'extract_waveforms', 'type':'group', 'children' : gui_params.waveforms_params},
            {'name':'clean_peaks', 'type':'group', 'children' : gui_params.clean_peaks_params},
            {'name':'noise_snippet', 'type':'group', 'children': gui_params.noise_snippet_params},
            {'name':'peak_sampler', 'type':'group', 'children' : gui_params.peak_sampler_params},
        ]        
        
        dia = ParamDialog(params_)
        d = {}
        for k in ('extract_waveforms', 'clean_peaks', 'noise_snippet', 'peak_sampler'):
            d[k] = self.catalogueconstructor.info[k]
        dia.set(d)

        
        dia.resize(450, 500)
        if dia.exec_():
            d = dia.get()
            #~ pprint(d)
            
            cc = self.catalogueconstructor
            
            cc.set_waveform_extractor_params(**d['extract_waveforms'])
            cc.clean_peaks(**d['clean_peaks'])
            cc.sample_some_peaks(**d['peak_sampler'])
            cc.extract_some_noise(**d['noise_snippet'])
            cc.compute_all_centroid(n_spike_for_centroid=_default_n_spike_for_centroid)
            self.refresh()

    def new_features(self):
        method, kargs = open_dialog_methods(gui_params.features_params_by_methods, self,
                    selected_method=self.catalogueconstructor.info['feature_method'])
        if method is not None:
            self.catalogueconstructor.extract_some_features(method=method, **kargs)
            self.refresh()


    def new_cluster(self):
        method, kargs = open_dialog_methods(gui_params.cluster_params_by_methods, self,
                    selected_method=self.catalogueconstructor.info['cluster_method'])
        if method is not None:
            self.catalogueconstructor.find_clusters(method=method, **kargs)
            self.controller.check_plot_attributes()
            self.refresh()
    
    def compute_metrics(self):
        dia = ParamDialog(gui_params.metrics_params)
        dia.resize(450, 500)
        if dia.exec_():
            d = dia.get()
            #~ self.catalogueconstructor.compute_spike_waveforms_similarity(method=d['spike_waveforms_similarity'], size_max=d['size_max'])
            self.catalogueconstructor.compute_cluster_similarity(method=d['cluster_similarity'])
            self.catalogueconstructor.compute_cluster_ratio_similarity(method=d['cluster_ratio_similarity'])
            self.catalogueconstructor.compute_spike_silhouette(size_max=d['size_max'])
            #TODO refresh only metrics concerned
            self.refresh()
        
    def export_peaks(self):
        #~ print('export_peaks')
        cc = self.catalogueconstructor
        export_catalogue_spikes(cc, export_path=None, formats=None)




"""



"""
try:
    import pyacq

    #test pyacq version
    import distutils.version
    assert distutils.version.LooseVersion(pyacq.__version__)>='0.2.0-dev'
    HAVE_PYACQ = True
except:
    HAVE_PYACQ = False


if HAVE_PYACQ:
    from .onlinepeeler import OnlinePeeler
    from .onlinetools import make_empty_catalogue, lighter_catalogue
    from .onlinetraceviewer import OnlineTraceViewer
    from .onlinewindow import TdcOnlineWindow
    from .onlinewaveformhistviewer import OnlineWaveformHistViewer
    from .launcher import make_pyacq_device_from_buffer, start_online_window, start_online_pyacq_buffer_demo, start_online_openephys
    from .probeactivityviewer import ProbeActivityViewer




from .myqt import QT
import pyqtgraph as pg

import numpy as np

from .. import labelcodes
from .base import WidgetBase
from .baselist import ClusterBaseList

from .tools import ParamDialog, open_dialog_methods
from . import gui_params


class PeakModel(QT.QAbstractItemModel):
    def __init__(self, parent =None, controller=None):
        QT.QAbstractItemModel.__init__(self,parent)
        self.controller = controller
        self.refresh_colors()
    
    def columnCount(self , parentIndex):
        return 6
        
    def rowCount(self, parentIndex):
        if not parentIndex.isValid() and self.controller.spike_label is not None:
            self.visible_ind, = np.nonzero(self.controller.spike_visible)
            n = self.visible_ind.size
            return n
        else :
            return 0
        
    def index(self, row, column, parentIndex):
        if not parentIndex.isValid():
            if column==0:
                childItem = row
            return self.createIndex(row, column, None)
        else:
            return QT.QModelIndex()
    
    def parent(self, index):
        return QT.QModelIndex()
    
    def data(self, index, role):
        if not index.isValid():
            return None
        
        if role not in (QT.Qt.DisplayRole, QT.Qt.DecorationRole):
            return
        
        col = index.column()
        row = index.row()
        #~ ind = self.visible_peak_labels.index[row]
        #~ label =  self.visible_peak_labels.iloc[row]
        #~ t_start = 0.
        
        abs_ind = self.visible_ind[row]
        
        seg_num = self.controller.spike_segment[abs_ind]
        peak_pos = self.controller.spike_index[abs_ind]
        peak_time = peak_pos/self.controller.dataio.sample_rate
        peak_label = self.controller.spike_label[abs_ind]
        peak_chan = self.controller.spike_channel[abs_ind]
        peak_ampl = self.controller.spikes[abs_ind]['extremum_amplitude']
        
        
        if role ==QT.Qt.DisplayRole :
            if col == 0:
                return '{}'.format(abs_ind)
            elif col == 1:
                return '{}'.format(seg_num)
            elif col == 2:
                return '{:.4f}'.format(peak_time)
            elif col == 3:
                return '{}'.format(peak_label)
            elif col == 4:
                return '{}'.format(peak_chan)
            elif col == 5:
                return '{:.1f}'.format(peak_ampl)
            else:
                return None
        elif role == QT.Qt.DecorationRole :
            if col != 0: return None
            if peak_label in self.icons:
                return self.icons[peak_label]
            else:
                return None
        else :
            return None
    
    def flags(self, index):
        if not index.isValid():
            return QT.Qt.NoItemFlags
        return QT.Qt.ItemIsEnabled | QT.Qt.ItemIsSelectable #| Qt.ItemIsDragEnabled

    def headerData(self, section, orientation, role):
        if orientation == QT.Qt.Horizontal and role == QT.Qt.DisplayRole:
            return  ['num', 'seg_num', 'time', 'cluster_label', 'channel', 'amplitude'][section]
        return
    
    def refresh_colors(self):
        self.icons = { }
        for k in self.controller.qcolors:
            color = self.controller.qcolors.get(k, QT.QColor( 'white'))
            pix = QT.QPixmap(10,10 )
            pix.fill(color)
            self.icons[k] = QT.QIcon(pix)
        
        #~ self.icons[-1] = QIcon(':/user-trash.png')
        
        self.layoutChanged.emit()
        
        
class PeakList(WidgetBase):
    """
    **Peak List** show all detected peak for the catalogue construction.
    
    Here pintentionally peaks are not spikes already (even most of them are spikes)
    because supperposition of spikes are done here in catalogue in Peeler.
    
    Note:
      * If there are to many peaks, not all of them will have a extracted waveform.
        This why some peak are not labeled (-10) and nb_peak != nb_wveforms
        sometimes.
      * Peaks can belong to diffrents segment, a column indicate it. This is th full list
        of all peaks of all segment.
      * A right click open a ontext menu:
        * move one or several selected spike to trash
        * create a new cluster with one or several spikes
      * When you select one spike, this will auto zoom on **Trace View**,  auto select
        the appriopriate segment and hilight the point on **ND Scatetr**. And vice versa.
    """
    def __init__(self, controller=None, parent=None):
        WidgetBase.__init__(self, parent=parent, controller=controller)
        
        self.layout = QT.QVBoxLayout()
        self.setLayout(self.layout)
        
        self.label_title = QT.QLabel('')
        self.layout.addWidget(self.label_title)
        
        self.tree = QT.QTreeView(minimumWidth = 100, uniformRowHeights = True,
                    selectionMode= QT.QAbstractItemView.ExtendedSelection, selectionBehavior = QT.QTreeView.SelectRows,
                    contextMenuPolicy = QT.Qt.CustomContextMenu,)
        
        self.layout.addWidget(self.tree)
        self.tree.customContextMenuRequested.connect(self.open_context_menu)
        
        self.model = PeakModel(controller = controller)
        self.tree.setModel(self.model)
        self.tree.selectionModel().selectionChanged.connect(self.on_tree_selection)

        # This is very slow!!!!!
        #~ for i in range(self.model.columnCount(None)):
            #~ print(i)
            #~ self.tree.resizeColumnToContents(i)
        self.tree.setColumnWidth(0,80)
        
        self.refresh()
    
    def refresh(self):
        self.model.refresh_colors()
        nb_peak = self.controller.spikes.size
        if self.controller.some_peaks_index is not None:
            nb_sampled = self.controller.some_peaks_index.shape[0]
        else:
            nb_sampled = 0
        self.label_title.setText('<b>All peaks {} - Nb sampled {}</b>'.format(nb_peak, nb_sampled))
    
    def on_tree_selection(self):
        self.controller.spike_selection[:] = False
        for index in self.tree.selectedIndexes():
            if index.column() == 0:
                ind = self.model.visible_ind[index.row()]
                self.controller.spike_selection[ind] = True
        self.spike_selection_changed.emit()
    
    def on_spike_selection_changed(self):
        self.tree.selectionModel().selectionChanged.disconnect(self.on_tree_selection)
        
        row_selected, = np.nonzero(self.controller.spike_selection[self.model.visible_ind])
        
        if row_selected.size>100:#otherwise this is verry slow
            row_selected = row_selected[:10]
        
        # change selection
        self.tree.selectionModel().clearSelection()
        flags = QT.QItemSelectionModel.Select #| QItemSelectionModel.Rows
        itemsSelection = QT.QItemSelection()
        for r in row_selected:
            for c in range(2):
                index = self.tree.model().index(r,c,QT.QModelIndex())
                ir = QT.QItemSelectionRange( index )
                itemsSelection.append(ir)
        self.tree.selectionModel().select(itemsSelection , flags)

        # set selection visible
        if len(row_selected)>=1:
            index = self.tree.model().index(row_selected[0],0,QT.QModelIndex())
            self.tree.scrollTo(index)

        self.tree.selectionModel().selectionChanged.connect(self.on_tree_selection)        


    def open_context_menu(self):
        menu = QT.QMenu()
        act = menu.addAction('Move peak selection to trash')
        act.triggered.connect(self.move_peak_selection_to_trash)
        act = menu.addAction('Make cluster with selection')
        act.triggered.connect(self.make_new_cluster)

        ##menu.popup(self.cursor().pos())
        menu.exec_(self.cursor().pos())
    
    def move_peak_selection_to_trash(self):
        self.controller.change_spike_label(self.controller.spike_selection, -1)
        self.refresh()
        self.spike_label_changed.emit()

    def make_new_cluster(self):
        self.controller.change_spike_label(self.controller.spike_selection, max(self.controller.cluster_labels)+1)
        self.refresh()
        self.spike_label_changed.emit()
    
    
class ClusterPeakList(ClusterBaseList):
    """
    **Cluster list** is the central widget for actions for clusters :
    make them visible, merge, trash, sort, split, change color, ...
    
    A context menu with right propose:
      * **Reset colors**
      * **Show all**
      * **Hide all**
      * **Re-label cluster by rms**: this re order cluster so that 0 have the bigger rms
         and N the lowest.
      * **Feature projection with all**: this re compute feature projection (equivalent to left toolbar)
      * **Feature projection with selection**: this re compute feature projection but take
        in account only selected usefull when you have doubt on small cluster and want a specifit
        PCA because variance is absord by big ones.
      * **Move selection to trash**
      * **Merge selection**: re label spikes in the same cluster.
      * **Select on peak list**: a spike  as selected for theses clusters.
      * **Tag selection as same cell**: in case of burst some cell
        can have diffrent waveform shape leading to diferents cluster but
        with same ratio. If you have that do not merge clusters because the
        Peeler wll fail. Prefer tag 2 cluster as same cell.
      * **Split selection**: try to split only selected cluster.
    
    Double click on a row make invisible all others except this one.
    
    Cluster can be visualy ordered by some criteria (rms, amplitude, nb peak, ...)
    This is useless to explore cluster few peaks, or big amplitudes, ...
    
    Negative labels are reserved:
      * -1 is thrash
      * -2 is noise snippet
      * -10 unclassified (because no waveform associated)
      * -9 Alien

    """
    _special_label = sorted(list(labelcodes.to_name.keys()))

    def make_menu(self):
        self.menu = QT.QMenu()

        act = self.menu.addAction('Reset colors')
        act.triggered.connect(self.reset_colors)
        act = self.menu.addAction('Show all')
        act.triggered.connect(self.show_all)
        act = self.menu.addAction('Hide all')
        act.triggered.connect(self.hide_all)
        act = self.menu.addAction('Re-label cluster by rms')
        act.triggered.connect(self.order_clusters)
        
        act = self.menu.addAction('Feature projection with all')
        act.triggered.connect(self.pc_project_all)
        act = self.menu.addAction('Feature projection with selection')
        act.triggered.connect(self.pc_project_selection)
        act = self.menu.addAction('Move cluster selection to trash')
        act.triggered.connect(self.move_cluster_selection_to_trash)
        act = self.menu.addAction('Merge selection')
        act.triggered.connect(self.merge_selection)
        act = self.menu.addAction('Select on peak list')
        act.triggered.connect(self.select_peaks_of_clusters)
        act = self.menu.addAction('Tag selected clusters as same cell')
        act.triggered.connect(self.selection_tag_same_cell)
        
        act = self.menu.addAction('Split selection')
        act.triggered.connect(self.split_selection)

        act = self.menu.addAction('Change color/annotation/tag')
        act.triggered.connect(self.change_color_annotation_tag)

    def _selected_spikes(self):
        selection = np.zeros(self.controller.spike_label.shape[0], dtype = bool)
        for k in self.selected_cluster():
            selection |= self.controller.spike_label == k
        return selection
    
    def reset_colors(self):
        self.controller.refresh_colors(reset = True)
        self.refresh()
        self.colors_changed.emit()
    
    def order_clusters(self):
        self.controller.order_clusters()
        self.refresh()
        self.spike_label_changed.emit()
    
    def pc_project_all(self, selection=None):
        
        params = gui_params.features_params_by_methods
        if selection is not None:
            params = params.copy()
            params['global_lda'] = []
        
        method, kargs = open_dialog_methods(params, self)
        
        if method is None:
            return
        
        self.controller.extract_some_features(method=method, selection=selection, **kargs)
        self.refresh()
        self.spike_label_changed.emit()
    
    def pc_project_selection(self):
        print('pc_project_selection', np.sum(self._selected_spikes()))
        self.pc_project_all(selection=self._selected_spikes())
    
    def move_cluster_selection_to_trash(self):
        mask = np.zeros(self.controller.spike_label.size, dtype='bool')
        for k in self.selected_cluster():
            #~ mask = self.controller.spike_label == k
            mask |= self.controller.spike_label == k
        self.controller.change_spike_label(mask, labelcodes.LABEL_TRASH)
        self.refresh()
        self.spike_label_changed.emit()
    
    def merge_selection(self):
        label_to_merge = self.selected_cluster()
        self.controller.merge_cluster(label_to_merge)
        self.refresh()
        self.spike_label_changed.emit()
    
    def split_selection(self):
        #TODO bug when not n_clusters
        
        n = len(self.selected_cluster())
        if n!=1: return
        label_to_split = self.selected_cluster()[0]
        
        method, kargs = open_dialog_methods(gui_params.cluster_params_by_methods, self)
        
        if method is None: return
        
        self.controller.split_cluster(label_to_split, method=method,  **kargs) #order_clusters=True,
        self.refresh()
        self.spike_label_changed.emit()

    def selection_tag_same_cell(self):
        labels_to_group = self.selected_cluster()
        self.controller.tag_same_cell(labels_to_group)
        self.refresh()
        self.cluster_tag_changed.emit()

    def select_peaks_of_clusters(self):
        self.controller.spike_selection[:] = self._selected_spikes()
        self.refresh()
        self.spike_selection_changed.emit()
    
    def change_color_annotation_tag(self):
        labels = self.selected_cluster()
        n = len(labels)
        if n!=1: return
        k = labels[0]
        clusters = self.controller.clusters
        ## ind = np.searchsorted(clusters['cluster_label'], k)  ## wrong because searchsortedmust be ordered
        ind = np.nonzero(clusters['cluster_label'] == k)[0][0]
        
        color = QT.QColor(self.controller.qcolors[k])
        annotations = str(clusters[ind]['annotations'])
        tag = str(clusters[ind]['tag'])
        
        params_ = [
            {'name': 'color', 'type': 'color', 'value': color},
            {'name': 'annotations', 'type': 'str', 'value': annotations},
            {'name': 'tag', 'type': 'list', 'value': tag, 'values':gui_params.possible_tags},
        ]        
        
        dia = ParamDialog(params_, title = 'Cluster {}'.format(k), parent=self)
        if dia.exec_():
            d = dia.get()
            self.controller.set_cluster_attributes(k, **d)
            
            self.colors_changed.emit()
            self.refresh()




from setuptools import setup
import os

d = {}
exec(open("tridesclous/version.py").read(), None, d)
version = d['version']

install_requires = [
                    'numpy<1.22',
                    'scipy',
                    'pandas',
                    'openpyxl',
                    'scikit-learn>=0.22.2',
                    'matplotlib',
                    'seaborn',
                    'neo>=0.8',
                    'tqdm',
                    # 'PyQt5',  make conda buggy
                    'pyqtgraph',
                    'joblib',
                    'numba',
                    'hdbscan',
                    'loky',
                    ]
extras_require={
                            'online' : ['pyacq',],
                            'opencl' : ['pyopencl'],
                        }

long_description = ""

setup(
    name = "tridesclous",
    version = version,
    packages = ['tridesclous', 'tridesclous.gui', 'tridesclous.gui.icons',
                'tridesclous.online', 'tridesclous.scripts', 'tridesclous.tests'],
    install_requires=install_requires,
    extras_require = extras_require,
    author = "C. Pouzat, S.Garcia",
    author_email = "",
    description = "offline/online spike sorting with french touch that light the barbecue",
    long_description = long_description,
    entry_points={
          'console_scripts': ['tdc=tridesclous.scripts.tdc:main'],
          #~ 'gui_scripts': ['tdcgui=tridesclous.scripts.tdc:open_mainwindow'],
        },
     license = "MIT",
    url='https://github.com/tridesclous/trisdesclous',
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering']
)



import os
import shutil
from pprint import pprint

import numpy as np

from tridesclous.datasets import download_dataset
from tridesclous.dataio import DataIO
from tridesclous.catalogueconstructor import CatalogueConstructor
from tridesclous.cataloguetools import apply_all_catalogue_steps
from tridesclous.autoparams import get_auto_params_for_catalogue

def is_running_on_ci_cloud():
    
    if os.environ.get('TRAVIS') in ('true', 'True'):
        return True
    
    if os.environ.get('APPVEYOR') in ('true', 'True'):
        return True

    if os.environ.get('CIRCLECI') in ('true', 'True'):
        return True
    
    return False

ON_CI_CLOUD = is_running_on_ci_cloud()
    
    

def setup_catalogue(dirname, dataset_name='olfactory_bulb', 
                duration=None, peak_sampler_mode=None):
    if os.path.exists(dirname):
        shutil.rmtree(dirname)
        
    dataio = DataIO(dirname=dirname)
    localdir, filenames, params = download_dataset(name=dataset_name)
    dataio.set_data_source(type='RawData', filenames=filenames, **params)
    
    if dataset_name=='olfactory_bulb':
        channels = [4, 5, 6, 7, 8, 9]
        mode = 'sparse'
        adjacency_radius_um = 350
        peak_method = 'geometrical'
        peak_engine = 'numpy'
        
        feature_method = 'pca_by_channel'
        feature_kargs = {'n_components_by_channel': 3}

        cluster_method = 'pruningshears'
        cluster_kargs = {'adjacency_radius_um' : 350 }
        
    else:
        channels = [0,1,2,3]
        mode = 'dense'
        adjacency_radius_um = None
        peak_method = 'global'
        peak_engine = 'numpy'
        
        feature_method = 'global_pca'
        feature_kargs = {'n_components': 6}

        cluster_method = 'pruningshears'
        cluster_kargs = {'adjacency_radius_um' : 150 }
        

    dataio.add_one_channel_group(channels=channels)
    
    
    cc = CatalogueConstructor(dataio=dataio)
    
    params = get_auto_params_for_catalogue(dataio)
    params['mode'] = mode
    params['n_jobs'] = 1
    params['peak_detector']['method'] = peak_method
    params['peak_detector']['engine'] = peak_engine
    params['peak_detector']['adjacency_radius_um'] = adjacency_radius_um
    params['feature_method'] = feature_method
    params['feature_kargs'] = feature_kargs
    params['cluster_method'] = cluster_method
    params['cluster_kargs'] = cluster_kargs
    
    if duration is not None:
        params['duration'] = duration
    
    if peak_sampler_mode is not None:
        params['peak_sampler']['mode'] = peak_sampler_mode
        
    
    #~ pprint(params)
    cc.apply_all_steps(params, verbose=True)
    
    # already done in apply_all_catalogue_steps:
    # cc.make_catalogue_for_peeler(inter_sample_oversampling=False, catalogue_name='initial') 
    
    # DEBUG
    #~ cc.make_catalogue_for_peeler(inter_sample_oversampling=True, catalogue_name='with_oversampling')
    
    return cc, params




if __name__ =='__main__':
    print('is_running_on_ci_cloud', is_running_on_ci_cloud())
    


import numpy as np
import pandas as pd

from .myqt import QT
import pyqtgraph as pg


from .peelercontroller import PeelerController
from .traceviewer import PeelerTraceViewer
from .spikelists import SpikeList, ClusterSpikeList
from .waveformviewer import PeelerWaveformViewer
from .waveformhistviewer import WaveformHistViewer
from .isiviewer import ISIViewer
from .crosscorrelogramviewer import CrossCorrelogramViewer

from . import icons

import itertools
import datetime



    

class PeelerWindow(QT.QMainWindow):
    def __init__(self, parent=None, dataio=None, catalogue=None):
        QT.QMainWindow.__init__(self, parent=None)
        
        self.setWindowIcon(QT.QIcon(':/main_icon.png'))
        
        self.controller = PeelerController(dataio=dataio, catalogue=catalogue)
        
        self.traceviewer = PeelerTraceViewer(controller=self.controller)
        self.spikelist = SpikeList(controller=self.controller)
        self.clusterlist = ClusterSpikeList(controller=self.controller)
        self.waveformviewer = PeelerWaveformViewer(controller=self.controller)
        self.waveformhistviewer = WaveformHistViewer(controller=self.controller)
        self.isiviewer = ISIViewer(controller=self.controller)
        self.crosscorrelogramviewer = CrossCorrelogramViewer(controller=self.controller)
        
        all = [self.traceviewer, self.spikelist, self.clusterlist, self.isiviewer]
        
        docks = {}


        docks['isiviewer'] = QT.QDockWidget('isiviewer',self)
        docks['isiviewer'].setWidget(self.isiviewer)
        self.addDockWidget(QT.Qt.RightDockWidgetArea, docks['isiviewer'])

        docks['crosscorrelogramviewer'] = QT.QDockWidget('crosscorrelogramviewer',self)
        docks['crosscorrelogramviewer'].setWidget(self.crosscorrelogramviewer)
        self.tabifyDockWidget(docks['isiviewer'], docks['crosscorrelogramviewer'])

        docks['waveformviewer'] = QT.QDockWidget('waveformviewer',self)
        docks['waveformviewer'].setWidget(self.waveformviewer)
        #~ self.addDockWidget(QT.Qt.RightDockWidgetArea, docks['waveformviewer'])
        self.tabifyDockWidget(docks['crosscorrelogramviewer'], docks['waveformviewer'])
        
        docks['waveformhistviewer'] = QT.QDockWidget('waveformhistviewer',self)
        docks['waveformhistviewer'].setWidget(self.waveformhistviewer)
        #~ self.addDockWidget(QT.Qt.RightDockWidgetArea, docks['waveformviewer'])
        self.tabifyDockWidget(docks['waveformviewer'], docks['waveformhistviewer'])
        
        
        docks['traceviewer'] = QT.QDockWidget('traceviewer',self)
        docks['traceviewer'].setWidget(self.traceviewer)
        #~ self.addDockWidget(QT.Qt.RightDockWidgetArea, docks['traceviewer'])
        self.tabifyDockWidget(docks['waveformhistviewer'], docks['traceviewer'])

        docks['spikelist'] = QT.QDockWidget('spikelist',self)
        docks['spikelist'].setWidget(self.spikelist)
        self.addDockWidget(QT.Qt.LeftDockWidgetArea, docks['spikelist'])

        docks['clusterlist'] = QT.QDockWidget('clusterlist',self)
        docks['clusterlist'].setWidget(self.clusterlist)
        self.splitDockWidget(docks['spikelist'], docks['clusterlist'], QT.Qt.Horizontal)





from tridesclous import *
from tridesclous.gui import GpuSelector
import  pyqtgraph as pg

from tridesclous.cltools import HAVE_PYOPENCL


import pytest

@pytest.mark.skip(not HAVE_PYOPENCL, reason='need opencl')
def test_GpuSelector():
    
    app = mkQApp()

    appname = 'tridesclous'
    settings_name = 'settings'
    settings = QT.QSettings(appname, settings_name)
    
    view = GpuSelector(settings=settings)
    
    
    if __name__ == '__main__':
        #~ app.exec_()
        if view.exec_():
            view.apply_cl_setting()
            d = view.get()
            
            print(d)

    


if __name__ == '__main__':
    test_GpuSelector()



from .myqt import QT
import pyqtgraph as pg

import time


class ControllerBase(QT.QObject):
    spike_selection_changed = QT.pyqtSignal()
    spike_label_changed = QT.pyqtSignal()
    colors_changed = QT.pyqtSignal()
    cluster_visibility_changed = QT.pyqtSignal()
    cluster_tag_changed = QT.pyqtSignal()
    
    def __init__(self, parent=None):
        QT.QObject.__init__(self, parent=parent)
        self.views = []
    
    def declare_a_view(self, new_view):
        assert new_view not in self.views, 'view already declared {}'.format(self)
        self.views.append(new_view)
        
        new_view.spike_selection_changed.connect(self.on_spike_selection_changed)
        new_view.spike_label_changed.connect(self.on_spike_label_changed)
        new_view.colors_changed.connect(self.on_colors_changed)
        new_view.cluster_visibility_changed.connect(self.on_cluster_visibility_changed)
        new_view.cluster_tag_changed.connect(self.on_cluster_tag_changed)
        
    def on_spike_selection_changed(self):
        for view in self.views:
            if view==self.sender(): continue
            #~ t1 = time.perf_counter()
            view.on_spike_selection_changed()
            #~ t2 = time.perf_counter()
            #~ print('on_spike_selection_changed',view,  t2-t1)

    def on_spike_label_changed(self):
        for view in self.views:
            if view==self.sender(): continue
            #~ t1 = time.perf_counter()
            view.on_spike_label_changed()
            #~ t2 = time.perf_counter()
            #~ print('on_spike_label_changed',view,  t2-t1)
    
    def on_colors_changed(self):
        for view in self.views:
            if view==self.sender(): continue
            #~ t1 = time.perf_counter()
            view.on_colors_changed()
            #~ t2 = time.perf_counter()
            #~ print('on_colors_changed',view,  t2-t1)
    
    def on_cluster_visibility_changed(self):
        #~ print('on_cluster_visibility_changed', self.cluster_visible)
        for view in self.views:
            if view==self.sender(): continue
            #~ t1 = time.perf_counter()
            view.on_cluster_visibility_changed()
            #~ t2 = time.perf_counter()
            #~ print('on_cluster_visibility_changed',view,  t2-t1)

    def on_cluster_tag_changed(self):
        for view in self.views:
            if view==self.sender(): continue
            #~ t1 = time.perf_counter()
            view.on_cluster_tag_changed()
            #~ t2 = time.perf_counter()
            #~ print('on_cluster_tag_changed',view,  t2-t1)

    @property
    def channel_indexes(self):
        channel_group = self.dataio.channel_groups[self.chan_grp]
        return channel_group['channels']

    @property
    def channel_names(self):
        all_names = self.dataio.datasource.get_channel_names()
        return [all_names[c] for c in self.channel_indexes]
    
    @property
    def channel_indexes_and_names(self):
        all_names = self.dataio.datasource.get_channel_names()
        return [ (c, all_names[c]) for c in self.channel_indexes]
    


class WidgetBase(QT.QWidget):
    spike_selection_changed = QT.pyqtSignal()
    spike_label_changed = QT.pyqtSignal()
    colors_changed = QT.pyqtSignal()
    cluster_visibility_changed = QT.pyqtSignal()
    cluster_tag_changed = QT.pyqtSignal()
    
    _params = None
    
    def __init__(self, parent = None, controller=None):
        QT.QWidget.__init__(self, parent)
        self.controller = controller
        if self.controller is not None:
            self.controller.declare_a_view(self)
        
        if self._params is not None:
            self.create_settings()
    
    def refresh(self):
        raise(NotImplementedError)

    def create_settings(self):
        self.params = pg.parametertree.Parameter.create( name='settings', type='group', children=self._params)
        
        self.tree_params = pg.parametertree.ParameterTree(parent=self)
        self.tree_params.header().hide()
        self.tree_params.setParameters(self.params, showTop=True)
        self.tree_params.setWindowTitle(u'Options for waveforms hist viewer')
        self.tree_params.setWindowFlags(QT.Qt.Window)
        
        self.params.sigTreeStateChanged.connect(self.on_params_changed)

    def open_settings(self):
        if not self.tree_params.isVisible():
            self.tree_params.show()
        else:
            self.tree_params.hide()
    
    def on_params_changed(self):
        self.refresh()
    
    def on_spike_selection_changed(self):
        self.refresh()

    def on_spike_label_changed(self):
        self.refresh()
        
    def on_colors_changed(self):
        self.refresh()
    
    def on_cluster_visibility_changed(self):
        self.refresh()
    
    def on_cluster_tag_changed(self):
        pass




import numpy as np
import sklearn.metrics.pairwise
import re
from urllib.request import urlretrieve
import zipfile
import os
import json

from . import labelcodes

def median_mad(data, axis=0):
    """
    Compute along axis the median and the mad.
    
    Arguments
    ----------------
    data : np.ndarray
    
    
    Returns
    -----------
    med: np.ndarray
    mad: np.ndarray
    
    
    """
    med = np.median(data, axis=axis)
    mad = np.median(np.abs(data-med),axis=axis)*1.4826
    return med, mad


def get_pairs_over_threshold(m, labels, threshold):
    """
    detect pairs over threhold in a similarity matrice
    """
    m = np.triu(m)
    ind0, ind1 = np.nonzero(m>threshold)
    
    #remove diag
    keep = ind0!=ind1
    ind0 = ind0[keep]
    ind1 = ind1[keep]
    
    pairs = list(zip(labels[ind0], labels[ind1]))
    
    return pairs
    


class FifoBuffer:
    """
    Kind of fifo on axis 0 than ensure to have the buffer and partial of previous buffer
    continuous in memory.
    
    This is not efficient if shape[0]is lot greater than chunksize.
    But if shape[0]=chunksize+smallsize, it should be OK.
    
    
    """
    def __init__(self, shape, dtype):
        self.buffer = np.zeros(shape, dtype=dtype)
        self.last_index = None
        
    def new_chunk(self, data, index):
        if self.last_index is not None:
            assert self.last_index+data.shape[0]==index, 'FifoBuffer self.last_index+data.shape[0]==index {} {}'.format(self.last_index+data.shape[0], index)
        
        assert data.shape[0]<=self.buffer.shape[0]
        n = self.buffer.shape[0]-data.shape[0]
        assert n>0
        #roll the end
        
        self.buffer[:n] = self.buffer[-n:]
        self.buffer[n:] = data
        self.last_index = index
    
    def get_data(self, start, stop):
        start = start - self.last_index + self.buffer .shape[0]
        stop = stop - self.last_index + self.buffer .shape[0]
        assert start>=0
        assert stop<=self.buffer.shape[0]
        return self.buffer[start:stop]
    
    def reset(self):
        self.last_index = None
        self.buffer[:] = 0


def get_neighborhood(geometry, radius_um):
    """
    get neighborhood given geometry array and radius
    
    params
    -----
    geometry: numpy array (nb_channel, 2) intresect units ar micro meter (um)
    
    radius_um: radius in micro meter
    
    returns
    ----
    
    neighborhood: boolean numpy array (nb_channel, nb_channel)
    
    """
    d = sklearn.metrics.pairwise.euclidean_distances(geometry)
    return d<=radius_um

def open_prb(probe_filename):
    d = {}
    with open(probe_filename) as f:
        exec(f.read(), None, d)
    channel_groups = d['channel_groups']
    return channel_groups

read_prb = open_prb

def create_prb_file_from_dict(channel_groups, filename):
    # transform array to list
    #~ channel_groups_ = {}
    #~ for chan_grp, channel_group in channel_groups.items():
        #~ channel_groups_[chan_grp] = dict(channel_group)
        #~ if isinstance(channel_group['channels'], np.ndarray):
            #~ channel_groups_[chan_grp]['channels'] = channel_group['channels'].tolist()
    
    # write with hack on json to put key as inteteger (normally not possible in json)
    with open(filename, 'w', encoding='utf8') as f:
        txt = json.dumps(channel_groups,indent=4)
        for chan_grp in channel_groups.keys():
            txt = txt.replace('"{}":'.format(chan_grp), '{}:'.format(chan_grp))
            for chan in channel_groups[chan_grp]['channels']:
                txt = txt.replace('"{}":'.format(chan), '{}:'.format(chan))
        txt = 'channel_groups = ' +txt
        f.write(txt)


def fix_prb_file_py2(probe_filename):
    """
    prb file can be define in python2
    unfortunatly some of them are done with python
    and this is not working in python3
    range(0, 17) + range(18, 128)
    
    This script tryp to change range(...) by list(range(...)))
    """
    with open(probe_filename, 'rb') as f:
        prb = f.read()
    

    pattern = b"list\(range\([^()]*\)\)"
    already_ok = re.findall( pattern, prb)
    if len(already_ok)>0:
        return
    #~ print(already_ok)


    pattern = b"range\([^()]*\)"
    changes = re.findall( pattern, prb)
    for change in changes:
        prb = prb.replace(change, b'list('+change+b')')

    with open(probe_filename, 'wb') as f:
        f.write(prb)


def construct_probe_list():
    #this download probes list from klusta and circus
    # and put them in a file
    urls = [ 
                ('kwikteam', 'https://codeload.github.com/kwikteam/probes/zip/master', 'probes-master/'),
                ('spyking-circus', 'https://codeload.github.com/spyking-circus/spyking-circus/zip/master', 'spyking-circus-master/probes/'),
    ]
    
    #download zip from github
    probes = {}
    for name, url, path in urls:
        zip_name = url.split('/')[-1]
        urlretrieve(url, name+'.zip')
        
        with zipfile.ZipFile( name+'.zip') as zfile:
            probes[name] = []
            for f in zfile.namelist():
                if f.startswith(path) and f != path and not f.endswith('/'):
                    probes[name].append(f.replace(path, '').replace('.prb', ''))
    
    #generate .py with probe list
    with open('probe_list.py', 'w+') as f:
        f.write('# This file is generated do not modify!!!\n')
        f.write('probe_list = {\n')
        for name, url, path in urls:
            f.write('    #########\n')
            f.write('    "{}" : [\n'.format(name))
            for probe in probes[name]:
                f.write('        "{}",\n'.format(probe))
            f.write('    ],\n')
        f.write('}\n')

def download_probe(local_dirname, probe_name, origin='kwikteam'):
    """
    Download prb file from either kwikteam or spyking-circus github.
    """
    if origin == 'kwikteam':
        #Max Hunter made a list of neuronexus probes, many thanks
        baseurl = 'https://raw.githubusercontent.com/kwikteam/probes/master/'
    elif origin == 'spyking-circus':
        # Pierre Yger made a list of various probe file, many thanks
        baseurl = 'https://raw.githubusercontent.com/spyking-circus/spyking-circus/master/probes/'
    else:
        raise(NotImplementedError)
    
    
    if not probe_name.endswith('.prb'):
        probe_name += '.prb'
    
    probe_filename = probe_name
    if '/' in probe_filename:
        probe_filename = probe_filename.split('/')[-1]
    probe_filename = os.path.join(local_dirname, probe_filename)
    
    urlretrieve(baseurl+probe_name, probe_filename)
    fix_prb_file_py2(probe_filename)#fix range to list(range
    
    return probe_filename
    

def compute_cross_correlograms(spike_indexes, spike_labels, 
                spike_segments, cluster_labels, sample_rate,
                window_size=0.1, bin_size=0.001, symmetrize=False,
                check_sorted=False):
    """
    Compute several cross-correlogram in one course
    from sevral cluster.
    
    This very elegant implementation is copy from phy package
    written by Cyril Rossant.
    
    
    Some sligh modification have been made to fit tdc datamodel because
    there are several segment handling in tdc.
    
    """
    assert sample_rate > 0.

    # Find `binsize`.
    bin_size = np.clip(bin_size, 1e-5, 1e5)  # in seconds
    binsize = int(sample_rate * bin_size)  # in samples
    assert binsize >= 1
    real_bin_size = binsize/sample_rate

    # Find `winsize_bins`.
    window_size = np.clip(window_size, 1e-5, 1e5)  # in seconds
    winsize_bins = 2 * int(.5 * window_size / bin_size) + 1

    assert winsize_bins >= 1
    assert winsize_bins % 2 == 1

    # Take the cluster oder into account.
    cluster_labels = cluster_labels.copy()
    cluster_labels.sort()
    
    # Like spike_labels, but with 0..n_clusters-1 indices.
    spike_labels_i = np.searchsorted(cluster_labels, spike_labels)
    
    cluster_labels = np.array(cluster_labels)
    n_clusters = cluster_labels.size
    correlograms = np.zeros((n_clusters, n_clusters, winsize_bins // 2 + 1), dtype='int32')
    
    nb_seg = np.max(spike_segments) + 1
    for seg_num in range(nb_seg):
    
        # Shift between the two copies of the spike trains.
        shift = 1

        
        keep = (spike_segments==seg_num) & np.in1d(spike_labels, cluster_labels)
        sp_indexes = spike_indexes[keep]
        sp_labels = spike_labels[keep]
        sp_segments = spike_segments[keep]
        sp_labels_i = spike_labels_i[keep]
        if check_sorted:
            # theprevent bugs if spike vector is not sorted
            # could append in peeler if some case (too bad!!!)
            order = np.argsort(sp_indexes)
            sp_indexes = sp_indexes[order]
            sp_labels = sp_labels[order]
            sp_segments = sp_segments[order] 
            sp_labels_i = sp_labels_i[order]

        # At a given shift, the mask precises which spikes have matching spikes
        # within the correlogram time window.
        mask = np.ones_like(sp_indexes, dtype=np.bool)

        # The loop continues as long as there is at least one spike with
        # a matching spike.
        while mask[:-shift].any():
            # Number of time samples between spike i and spike i+shift.
            #~ spike_diff = _diff_shifted(spike_indexes, shift)
            spike_diff = sp_indexes[shift:] - sp_indexes[:len(sp_indexes) - shift]
            
            # Binarize the delays between spike i and spike i+shift.
            spike_diff_b = spike_diff // binsize

            # Spikes with no matching spikes are masked.
            mask[:-shift][spike_diff_b > (winsize_bins // 2)] = False

            # Cache the masked spike delays.
            m = mask[:-shift].copy()
            d = spike_diff_b[m]

            # Find the indices in the raveled correlograms array that need
            # to be incremented, taking into account the spike clusters.
            indices = np.ravel_multi_index((sp_labels_i[:-shift][m],
                                            sp_labels_i[+shift:][m],
                                            d),
                                           correlograms.shape)

            # Increment the matching spikes in the correlograms array.
            bbins = np.bincount(indices)
            correlograms.ravel()[:len(bbins)] += bbins
            
            
            shift += 1

        # Remove ACG peaks.
        correlograms[np.arange(n_clusters),
                 np.arange(n_clusters),
                 0] = 0
        
    if symmetrize:
        # We symmetrize c[i, j, 0].
        # This is necessary because the algorithm in correlograms()
        # is sensitive to the order of identical spikes.
        correlograms[..., 0] = np.maximum(correlograms[..., 0],
                                          correlograms[..., 0].T)
        sym = correlograms[..., 1:][..., ::-1]
        sym = np.transpose(sym, (1, 0, 2))
        correlograms = np.dstack((sym, correlograms))
        
        bins = np.arange(correlograms.shape[2]+1)*real_bin_size - real_bin_size*winsize_bins/2.

    else:
        bins = np.arange(correlograms.shape[2]+1)*real_bin_size - real_bin_size/2.
    
    return correlograms, bins
    


def get_color_palette(n, palette='husl', output='int32'):
    # this depend now on seaborn but seaborn will be renmoved as dependency soon
    # because it break joblib
    import seaborn as sns
    
    if output == 'rgb':
        colors = sns.color_palette(palette, n)
        return colors
    elif output == 'int32':
        colors_int32 = np.array([rgba_to_int32(r,g,b) for r,g,b in sns.color_palette(palette, n)])
        return colors_int32
    else:
        raise NotImplementedError


def int32_to_rgba(v, mode='int'):
    r = (v>>24) & 0xFF
    g = (v>>16) & 0xFF
    b = (v>>8) & 0xFF
    a = (v>>0) & 0xFF
    if mode == 'int':
        return r, g, b, a
    elif mode=='float':
        return r/255., g/255., b/255., a/255.


def rgba_to_int32(r, g, b, a=None):
    if type(r) == int:
        if a is None:
            a = 255
        #ensure max255
        r, g, b = (r & 0xFF), (g & 0xFF), (b & 0xFF)
        v = (r<<24) + (g<<16) + (b<<8) + a
    else:
        if a is None:
            a = 1.
        v = (int(r*255.)<<24) + (int(g*255.)<<16) + (int(b*255.)<<8) + int(a*255.)
    return v
        
def make_color_dict(clusters):
    colors = {}
    for cluster in clusters:
        r, g, b, a = int32_to_rgba(cluster['color'], mode='float')
        colors[cluster['cluster_label']] =  (r, g, b)
    colors[labelcodes.LABEL_TRASH] = (.4, .4, .4)
    colors[labelcodes.LABEL_UNCLASSIFIED] = (.6, .6, .6)
    colors[labelcodes.LABEL_NOISE] = (.8, .8, .8)
    colors[labelcodes.LABEL_ALIEN] = (.4, .8, .1)
    colors[labelcodes.LABEL_NO_WAVEFORM] = (.6, .6, .6)
    
    
    return colors



#~ if __name__=='__main__':
    #~ construct_probe_list()




import numpy as np
import time
import os
import shutil
from pprint import pprint

from tridesclous import download_dataset
from tridesclous.dataio import DataIO
from tridesclous.catalogueconstructor import CatalogueConstructor
from tridesclous.autoparams import get_auto_params_for_catalogue, get_auto_params_for_peelers

from matplotlib import pyplot as plt

def test_get_auto_params():
    if os.path.exists('test_autoparams'):
        shutil.rmtree('test_autoparams')
        
    dataio = DataIO(dirname='test_autoparams')
    #~ localdir, filenames, params = download_dataset(name='olfactory_bulb')
    localdir, filenames, params = download_dataset(name='locust')
    dataio.set_data_source(type='RawData', filenames=filenames, **params)
    
    params = get_auto_params_for_catalogue(dataio)
    pprint(params)
    
    
    params = get_auto_params_for_peelers(dataio)
    pprint(params)
    
    
    #~ print(params['cluster_method'])
    #~ print(params['cluster_kargs'])


if __name__ == '__main__':
    test_get_auto_params()



from .myqt import QT
import pyqtgraph as pg

import time
import os
from collections import OrderedDict
import pickle
import webbrowser
from pprint import pprint

from ..dataio import DataIO
from ..datasource import data_source_classes
from .tools import get_dict_from_group_param, ParamDialog, MethodDialog #, open_dialog_methods
from  ..datasets import datasets_info, download_dataset

from ..catalogueconstructor import CatalogueConstructor
from ..cataloguetools import apply_all_catalogue_steps
from ..autoparams import get_auto_params_for_catalogue, get_auto_params_for_peelers
from .cataloguewindow import CatalogueWindow
from ..peeler import Peeler
from .peelerwindow import PeelerWindow
from .initializedatasetwindow import InitializeDatasetWindow
from .probegeometryview import ProbeGeometryView
from ..export import export_list
from ..tools import open_prb
from ..report import generate_report
from ..cltools import HAVE_PYOPENCL
from .gpuselector import GpuSelector

from . import icons

from . import gui_params

try:
    import ephyviewer
    HAVE_EPHYVIEWER = True
except:
    HAVE_EPHYVIEWER = False


error_box_msg = """
This is the raw Python error.
Unfortunatly, this is sometimes not so usefull for user...
Please send it to https://github.com/tridesclous/tridesclous/issues.

{}
"""
class MainWindow(QT.QMainWindow):
    def __init__(self, parent=None):
        QT.QMainWindow.__init__(self, parent=parent)
        
        self.setWindowIcon(QT.QIcon(':/main_icon.png'))
        
        self.dataio = None
        #~ self.catalogueconstructor = None
        
        self.resize(800, 800)

        appname = 'tridesclous'
        settings_name = 'settings'
        self.settings = QT.QSettings(appname, settings_name)
        
        self.create_actions_and_menu()
        
        w = QT.QWidget()
        self.setCentralWidget(w)
        mainlayout  = QT.QVBoxLayout()
        w.setLayout(mainlayout)
        
        self.label_info = QT.QLabel('Nothing loaded')
        mainlayout.addWidget(self.label_info)
        
        mainlayout.addStretch()
        
        self.open_windows = []
        
        self.win_viewer = None
        self.probe_viewer = None
        
        self.dialog_fullchain_params = ParamDialog(gui_params.fullchain_params, parent=self)
        self.dialog_method_features = MethodDialog(gui_params.features_params_by_methods, parent=self,
                        title='Which feature method ?', selected_method='peak_max')
        self.dialog_method_cluster = MethodDialog(gui_params.cluster_params_by_methods, parent=self,
                        title='Which cluster method ?', selected_method = 'sawchaincut')
        
        self.dialog_method_peeler = MethodDialog(gui_params.peeler_params_by_methods, parent=self,
                    title='Peeler engine')
        
        if HAVE_PYOPENCL:
            self.dialog_gpuselector = GpuSelector(settings=self.settings)
    

    def create_actions_and_menu(self):
        
        self.toolbar = QT.QToolBar(orientation=QT.Vertical)
        self.toolbar.setToolButtonStyle(QT.Qt.ToolButtonTextUnderIcon)
        self.addToolBar(QT.TopToolBarArea, self.toolbar)
        self.toolbar.setIconSize(QT.QSize(60, 40))

        self.toolbar2 = QT.QToolBar(orientation=QT.Vertical)
        self.toolbar2.setToolButtonStyle(QT.Qt.ToolButtonTextUnderIcon)
        self.addToolBar(QT.LeftToolBarArea, self.toolbar2)
        self.toolbar2.setIconSize(QT.QSize(60, 40))
        
        
        self.file_menu = self.menuBar().addMenu(self.tr("File"))
        
        do_open = QT.QAction('&Open', self, shortcut = "Ctrl+O", icon=QT.QIcon(":document-open.svg"))
        do_open.triggered.connect(self.open_dialog)
        self.file_menu.addAction(do_open)
        self.toolbar.addAction(do_open)

        do_init = QT.QAction('&Initialize dataset', self, shortcut = "Ctrl+I", icon=QT.QIcon(":document-new.svg"))
        do_init.triggered.connect(self.initialize_dataset_dialog)
        self.file_menu.addAction(do_init)
        self.toolbar.addAction(do_init)
        
        self.recetly_opened_menu = self.file_menu.addMenu('Recently opened')
        self._refresh_recetly_opened()
        
        self.dw_dataset_menu = self.file_menu.addMenu('Download test datasets')
        for name in datasets_info:
            act = self.dw_dataset_menu.addAction(name)
            act.name = name
            act.triggered.connect(self.do_download_dataset)        
        
        
        self.toolbar.addSeparator()
        
        if HAVE_EPHYVIEWER:
            open_viewer = QT.QAction('&Preview raw signals', self, icon=QT.QIcon(":ephyviewer.png"))
            open_viewer.triggered.connect(self.open_ephyviewer)
            self.toolbar.addAction(open_viewer)
        
        open_probe_viewer = QT.QAction('&Probe geometry', self, icon=QT.QIcon(":probe-geometry.svg"))
        open_probe_viewer.triggered.connect(self.open_probe_viewer)
        self.toolbar.addAction(open_probe_viewer)
        
        self.toolbar.addSeparator()
        
        do_check_prb = QT.QAction('Check PRB file', self, ) # icon=QT.QIcon(":document-export.svg"))
        do_check_prb.triggered.connect(self.check_prb_file)
        self.file_menu.addAction(do_check_prb)
        
        if HAVE_PYOPENCL:
            do_open_gpuselector = QT.QAction('Default OpenCL (GPU) selector', self)
            do_open_gpuselector.triggered.connect(self.open_gpuselector)
            self.file_menu.addAction(do_open_gpuselector)
        
        self.toolbar.addSeparator()

        do_refresh = QT.QAction(u'Refresh', self,checkable = False, icon=QT.QIcon(":/view-refresh.svg"))
        do_refresh.triggered.connect(self.refresh_with_reload)
        self.toolbar.addAction(do_refresh)
        
        do_export_spikes = QT.QAction('Export spikes', self,  icon=QT.QIcon(":document-export.svg"))
        do_export_spikes.triggered.connect(self.export_spikes)
        self.toolbar.addAction(do_export_spikes)
        self.file_menu.addAction(do_export_spikes)

        do_generate_report = QT.QAction('Generate report', self,  icon=QT.QIcon(":kwrite.svg"))
        do_generate_report.triggered.connect(self.generate_report)
        self.toolbar.addAction(do_generate_report)
        self.file_menu.addAction(do_generate_report)

        self.toolbar.addSeparator()

        help_act = QT.QAction('Help', self,checkable = False, icon=QT.QIcon(":main_icon.png"))
        help_act.triggered.connect(self.open_webbrowser_help)
        self.toolbar.addAction(help_act)
        
        
        # Toolbar2
        self.toolbar2.addWidget(QT.QLabel('<b>Steps</b>'))
        
        self.toolbar2.addWidget(QT.QLabel('Select chanel group:'))
        self.combo_chan_grp = QT.QComboBox()
        self.toolbar2.addWidget(self.combo_chan_grp)
        self.combo_chan_grp.currentIndexChanged .connect(self.on_chan_grp_change)

        do_init_cataloguewin = QT.QAction('Initialize Catalogue', self, icon=QT.QIcon(":autocorrection.svg"))
        do_init_cataloguewin.triggered.connect(self.initialize_catalogue)
        self.toolbar2.addAction(do_init_cataloguewin)
        
        do_open_cataloguewin = QT.QAction('Open CatalogueWindow', self,  icon=QT.QIcon(":catalogwinodw.png"))
        do_open_cataloguewin.triggered.connect(self.open_cataloguewin)
        self.toolbar2.addAction(do_open_cataloguewin)


        do_run_peeler = QT.QAction('Run Peeler', self,  icon=QT.QIcon(":configure-shortcuts.svg"))
        do_run_peeler.triggered.connect(self.run_peeler)
        self.toolbar2.addAction(do_run_peeler)
        
        do_open_peelerwin = QT.QAction('Open PeelerWindow', self,  icon=QT.QIcon(":peelerwindow.png"))
        do_open_peelerwin.triggered.connect(self.open_peelerwin)
        self.toolbar2.addAction(do_open_peelerwin)

        
        
        
    def open_webbrowser_help(self):
        url = "http://tridesclous.readthedocs.io"
        webbrowser.open(url, new=2)
    
    def warn(self, text, title='Error in tridesclous'):
        mb = QT.QMessageBox.warning(self, title,text, 
                QT.QMessageBox.Ok ,
                QT.QMessageBox.NoButton)
    
    def errorToMessageBox(self, e):
        self.warn(error_box_msg.format(e))

    def refresh_info(self):
        txt = self.dataio.__repr__()
        txt += '\n\n'
        #~ print('refresh_info', self.chan_grps)
        for chan_grp in self.chan_grps:
            catalogueconstructor = CatalogueConstructor(dataio=self.dataio, chan_grp=chan_grp)
            txt += catalogueconstructor.__repr__()
            txt += '\n'
        
        self.label_info.setText(txt)
    
    def open_dialog(self):
        fd = QT.QFileDialog(fileMode=QT.QFileDialog.DirectoryOnly, acceptMode=QT.QFileDialog.AcceptOpen)
        #~ fd.setNameFilters(['Hearingloss setup (*.json)', 'All (*)'])
        fd.setViewMode( QT.QFileDialog.Detail )
        if fd.exec_():
            dirname = fd.selectedFiles()[0]
            print(dirname)
            
            if DataIO.check_initialized(dirname):
                self._open_dataio(dirname)
    
    def _refresh_recetly_opened(self):
        self.recetly_opened_menu.clear()
        for dirname in self.recently_opened():
            act = self.recetly_opened_menu.addAction(str(dirname))
            act.dirname = dirname
            act.triggered.connect(self.do_open_recent)
            
    def do_open_recent(self):
        self._open_dataio(self.sender().dirname)
    
    def recently_opened(self):
        value = self.settings.value('recently_opened')
        if value is None:
            recently_opened = []
        else:
            try:
                if type(value) == str:
                    #bug on some qt
                    value = value.encode('ascii')
                recently_opened = pickle.loads(value)
                if type(recently_opened) != list:
                    recently_opened = []
            except:
                recently_opened = []
        return recently_opened
    
    def do_download_dataset(self):
        name = self.sender().name
        fd = QT.QFileDialog(fileMode=QT.QFileDialog.DirectoryOnly, acceptMode=QT.QFileDialog.AcceptOpen)
        fd.setViewMode( QT.QFileDialog.Detail )
        if fd.exec_():
            localdir = os.path.join(fd.selectedFiles()[0], name)
            if os.path.exists(localdir):
                return
            os.mkdir(localdir)
            localdir, filenames, params = download_dataset(name, localdir=localdir)
            
            dirname = os.path.join(localdir, 'tdc_'+name)
            dataio = DataIO(dirname=dirname)
            dataio.set_data_source(type='RawData', filenames=filenames, **params)
            
            self._open_dataio( dirname)
    
    
    def refresh_with_reload(self):
        if self.dataio is None:
            return
        
        self.dataio = DataIO(dirname=self.dataio.dirname)
        self.refresh_info()
    
    def _open_dataio(self, dirname):
        recently_opened =self.recently_opened()
        if dirname not in recently_opened:
            recently_opened = [dirname] + recently_opened
            recently_opened = recently_opened[:5]
            self.settings.setValue('recently_opened', pickle.dumps(recently_opened))
            self._refresh_recetly_opened()
        
        for win in self.open_windows:
            win.close()
        self.open_windows = []
        
        self.dataio = DataIO(dirname=dirname)
        
        self.combo_chan_grp.blockSignals(True)
        self.combo_chan_grp.clear()
        self.combo_chan_grp.addItems([str(k) for k in self.dataio.channel_groups.keys()] + ['ALL'])
        self.combo_chan_grp.blockSignals(False)
        self.on_chan_grp_change()
        
        # set auto params catalogue
        params = get_auto_params_for_catalogue(self.dataio, chan_grp=self.chan_grps[0])
        d = dict(params)
        for k in ('feature_method', 'feature_kargs', 'cluster_method', 'cluster_kargs'):
            d.pop(k)
        self.dialog_fullchain_params.set(d)
        self.dialog_method_features.set_method(params['feature_method'], params['feature_kargs'])
        self.dialog_method_cluster.set_method(params['cluster_method'], params['cluster_kargs'])
        
        # TODO clean_cluster
        
        # set auto params peeler
        params = get_auto_params_for_peelers(self.dataio, chan_grp=self.chan_grps[0])
        d = dict(params)
        engine = d.pop('engine')
        self.dialog_method_peeler.set_method(engine, d)
        
        

    
    def release_closed_windows(self):
        for win in list(self.open_windows):
            if not win.isVisible():
                self.open_windows.remove(win)
    
    def close_window_chan_grp(self, chan_grp):
        for win in list(self.open_windows):
            if win.controller.chan_grp == chan_grp:
                win.close()
                self.open_windows.remove(win)
        

    #~ @property
    #~ def chan_grp(self):
        #~ txt = self.combo_chan_grp.currentText()
        #~ if txt == 'ALL':
            #~ # take the first
            #~ chan_grp = list(self.dataio.channel_groups.keys())[0]
        #~ else:
            #~ chan_grp = int(txt)
        #~ return chan_grp
    
    @property
    def chan_grps(self):
        txt = self.combo_chan_grp.currentText()
        if txt == 'ALL':
            # take the first
            chan_grps = list(self.dataio.channel_groups.keys())
        else:
            chan_grps = [int(txt)]
        return chan_grps
    
    def on_chan_grp_change(self, index=None):
        #~ self.catalogueconstructor = CatalogueConstructor(dataio=self.dataio, chan_grp=self.chan_grp)
        self.refresh_info()
        
        # this set a the a by default method depending the number of channels
        n = self.dataio.nb_channel(chan_grp=self.chan_grps[0])
        if 1<=n<9:
            feat_method = 'global_pca'
        else:
            feat_method = 'peak_max'
        self.dialog_method_features.param_method['method'] = feat_method
    
    def initialize_dataset_dialog(self):
        init_dia = InitializeDatasetWindow(parent=self)
        if init_dia.exec_():
            self._open_dataio(init_dia.dirname_created)


    
    def initialize_catalogue(self):
        
        if  self.dataio is None:
            return
        
        self.release_closed_windows()
        for chan_grp in self.chan_grps:
            self.close_window_chan_grp(chan_grp)
        
        self.refresh_with_reload()
        
        if not self.dialog_fullchain_params.exec_():
            return
        if not self.dialog_method_features.exec_():
            return
        if not self.dialog_method_cluster.exec_():
            return
        
        params = {}
        params.update(self.dialog_fullchain_params.get())
        params['memory_mode'] = 'memmap'
        #~ pprint(params)
        
        method, d = self.dialog_method_features.get()
        params['feature_method'] = method
        params['feature_kargs'] = d
        
        method, d = self.dialog_method_cluster.get()
        params['cluster_method'] = method
        params['cluster_kargs'] = d
        
        #~ #TODO dialog for that
        #~ params['clean_cluster'] = False
        #~ params['clean_cluster_kargs'] = {}
        

        for chan_grp in self.chan_grps:
            print('### chan_grp', chan_grp, ' ###')
        
            #~ try:
            if 1:
                catalogueconstructor = CatalogueConstructor(dataio=self.dataio, chan_grp=chan_grp)
                apply_all_catalogue_steps(catalogueconstructor, params, verbose=True)            
                
            #~ except Exception as e:
                #~ print(e)
                #~ self.errorToMessageBox(e)
                
        self.refresh_info()
    
    
    def open_cataloguewin(self):
        if self.dataio is None: return
        if len(self.chan_grps) != 1: return
            
        #~ try:
        if True:
            catalogueconstructor = CatalogueConstructor(dataio=self.dataio, chan_grp=self.chan_grps[0])
            win = CatalogueWindow(catalogueconstructor)
            win.setWindowTitle(self.dataio.channel_group_label(chan_grp=self.chan_grps[0]))
            win.show()
            self.open_windows.append(win)
        #~ except Exception as e:
            #~ print(e)
            #~ self.errorToMessageBox(e)
    
    def run_peeler(self):
        if self.dataio is None: return
        
        #TODO find something better when several segment
        #~ lengths = [ self.dataio.datasource.get_segment_shape(i)[0] for i in range(self.dataio.nb_segment)]
        #~ max_duration = max(lengths)/self.dataio.sample_rate
        
        total_duration = sum(self.dataio.get_segment_length(seg_num=seg_num) / self.dataio.sample_rate for seg_num in range(self.dataio.nb_segment))
        
        for m in self.dialog_method_peeler.methods:
            self.dialog_method_peeler.all_params[m]['duration'] = total_duration
        
        
        #~ dia = ParamDialog(gui_params.peeler_params)
        #~ dia.resize(450, 500)
        
        #~ if not dia.exec_():
            #~ return
        #~ d = dia.get()
            
        if not self.dialog_method_peeler.exec_():
            return
        engine, d = self.dialog_method_peeler.get()
        #~ print('run_peeler')
        #~ print(engine)
        #~ print(d)
        d['engine'] = engine

        #~ duration = d['duration'] if d['limit_duration'] else None
        #~ d.pop('limit_duration')
        duration = d.pop('duration')
        print('duration', duration)
        
        errors = []
        for chan_grp in self.chan_grps:
            try:
            #~ if True:
                initial_catalogue = self.dataio.load_catalogue(chan_grp=chan_grp)
                if initial_catalogue is None:
                    txt =  """chan_grp{}
Catalogue do not exists, please do:
    1. Initialize Catalogue (if not done)
    2. Open CatalogueWindow
    3. Make catalogue for peeler
                    """.format(chan_grp)
                    errors.append(txt)
                    continue
                
                pprint(d)
                peeler = Peeler(self.dataio)
                peeler.change_params(catalogue=initial_catalogue, **d)
                
                t1 = time.perf_counter()
                peeler.run(duration=duration)
                t2 = time.perf_counter()
                print('peeler.run_loop', t2-t1)
                
            except Exception as e:
                print(e)
                error = """chan_grp{}\n{}""".format(chan_grp, e)
                errors.append(error)
        
        for error in errors:
            self.errorToMessageBox(error)
    
    def open_peelerwin(self):
        if self.dataio is None: return
        
        if len(self.chan_grps) != 1:
            return
        
        try:
            initial_catalogue = self.dataio.load_catalogue(chan_grp=self.chan_grps[0])
            win = PeelerWindow(dataio=self.dataio, catalogue=initial_catalogue)
            win.show()
            self.open_windows.append(win)
        except Exception as e:
            print(e)
            self.errorToMessageBox(e)

    def open_ephyviewer(self):
        if self.win_viewer is not None:
            self.win_viewer.close()
            self.win_viewer = None
        
        if self.dataio  is None:
            return
        if not hasattr(self.dataio.datasource, 'rawios'):
            return
        
        sources = ephyviewer.get_sources_from_neo_rawio(self.dataio.datasource.rawios[0])
        
        self.win_viewer = ephyviewer.MainViewer()
        
        for i, sig_source in enumerate(sources['signal']):
            view = ephyviewer.TraceViewer(source=sig_source, name='signal {}'.format(i))
            view.params['scale_mode'] = 'same_for_all'
            view.params['display_labels'] = True
            view.auto_scale()
            if i==0:
                self.win_viewer.add_view(view)
            else:
                self.win_viewer.add_view(view, tabify_with='signal {}'.format(i-1))
        
        self.win_viewer.show()

    def open_probe_viewer(self):
        if self.dataio is None:
            return
        
        if self.probe_viewer is not None:
            self.probe_viewer.close()
            self.probe_viewer = None
        
        self.probe_viewer = ProbeGeometryView(channel_groups=self.dataio.channel_groups, parent=self)
        self.probe_viewer.setWindowFlags(QT.Qt.Window)
        self.probe_viewer.show()
    
    def export_spikes(self):
        if self.dataio is None:
            return
        
        possible_formats = [e.ext for e in export_list]
        params = [
            {'name': 'format', 'type': 'list', 'values':possible_formats},
            {'name': 'split_by_cluster', 'type': 'bool', 'value':False},
            {'name': 'use_cell_label', 'type': 'bool', 'value':True},
        ]
        dialog = ParamDialog(params, parent=self)
        if not dialog.exec_():
            return
        p = dialog.get()
        self.dataio.export_spikes(export_path=None,
                split_by_cluster=p['split_by_cluster'],  use_cell_label=p['use_cell_label'], formats=p['format'])

    def generate_report(self):
        if self.dataio is None:
            return
        generate_report(self.dataio)

    def check_prb_file(self):
        fd = QT.QFileDialog(fileMode=QT.QFileDialog.AnyFile, acceptMode=QT.QFileDialog.AcceptOpen)
        fd.setNameFilters(['Probe geometry (*.PRB, *.prb)', 'All (*)'])
        fd.setViewMode( QT.QFileDialog.Detail )
        if fd.exec_():
            prb_filename = fd.selectedFiles()[0]
            #~ print(prb_filename)
            try:
                channel_groups = open_prb(prb_filename)
                self.probe_viewer = ProbeGeometryView(channel_groups=channel_groups, parent=self)
            except:
                self.warn('Error in PRB')
                return
            self.probe_viewer.setWindowFlags(QT.Qt.Window)
            self.probe_viewer.show()
    
    def open_gpuselector(self):
        if not HAVE_PYOPENCL:
            return
        if not self.dialog_gpuselector.exec_():
            return
        self.dialog_gpuselector.apply_cl_setting()




import numpy as np
import os
import sys
import inspect
from urllib.request import urlretrieve


# For now, all testing file are in raw binary format
datasets_info = {
    'locust':{
        'url': 'https://raw.githubusercontent.com/tridesclous/tridesclous_datasets/master/locust/',
        'filenames': ['locust_trial_01.raw', 'locust_trial_02.raw'],
        'total_channel': 4,
        'dtype': 'int16',
        'sample_rate': 15000.,
        'bit_to_microVolt' : None, # TODO find it
    },
    
    'olfactory_bulb':{
        'url': 'https://raw.githubusercontent.com/tridesclous/tridesclous_datasets/master/olfactory_bulb/',
        'filenames': ['OB_file1.raw', 'OB_file2.raw', 'OB_file3.raw'],
        'total_channel': 16,
        'dtype': 'int16',
        'sample_rate': 10000.,
        'channel_group': list(range(14)),
        'bit_to_microVolt' : 20. * 1000./2**16, # This is not sure, but the order of magnitude is OK.
    },

    'purkinje':{
        'url': 'https://raw.githubusercontent.com/tridesclous/tridesclous_datasets/master/purkinje/',
        'filenames': ['purkinje_extra_cellular.raw'],
        'total_channel': 4,
        'dtype': 'float32',
        'sample_rate': 15000.,
        'bit_to_microVolt' : None, # TODO find it
    },
    'striatum_rat': {
        'url': 'https://raw.githubusercontent.com/tridesclous/tridesclous_datasets/master/striatum_rat/',
        'filenames': ['tetrode_striatum.dat'],
        'total_channel': 4,
        'dtype': 'int16',
        'sample_rate': 20000.,
        'prb_filename' : 'tetrode_striatum.prb',
        'bit_to_microVolt' : 0.195, # This is not sure, must be checked
    },
    
    
}


def download_dataset(name='locust', localdir=None):
    assert name in datasets_info
    
    if localdir is None:
        # get path of the calling function see
        #http://stackoverflow.com/questions/11757801/get-the-file-of-the-function-one-level-up-in-the-stack
        localdir = os.path.abspath(os.path.dirname(inspect.getfile(sys._getframe(1))))
        if not os.access(localdir, os.W_OK):
            localdir = tempfile.gettempdir()
        localdir = os.path.join(localdir, name)
    
    if not os.path.exists(localdir):
        os.mkdir(localdir)
    
    info = datasets_info[name]
    
    filenames = info['filenames']
    if 'prb_filename' in info:
        filenames = filenames + [ info['prb_filename'] ]
    
    for filename in filenames:
        #~ print(filename)
        localfile = os.path.join(localdir, filename)
        
        if not os.path.exists(localfile):
            distantfile = info['url'] + filename
            print('download_dataset', localfile, 'from', distantfile)
            urlretrieve(distantfile, localfile)
    
    filenames = [os.path.join(localdir, f) for f in info['filenames']]
    params = {k: datasets_info[name][k] for k in ('dtype', 'sample_rate', 'total_channel', 'bit_to_microVolt')}
    return localdir, filenames, params


def get_dataset(name='locust', localdir=None, seg_num=0):
    assert name in datasets_info
    
    localdir, filenames, params = download_dataset(name=name, localdir=localdir)
    filename = filenames[seg_num]
    data = np.memmap(os.path.join(localdir, filename), dtype=params['dtype'])
    
    data = data.reshape(-1, params['total_channel'])
    
    info = datasets_info[name]
    
    if 'channel_group' in info:
        data = data[:, datasets_info[name]['channel_group']]
    
    return data, params['sample_rate']
    




import pytest
import os, tempfile, shutil
import numpy as np

from tridesclous import download_dataset
from tridesclous import DataIO





    

def test_DataIO():
    
    
    # initialze dataio
    if os.path.exists('test_DataIO'):
        shutil.rmtree('test_DataIO')
        
    dataio = DataIO(dirname='test_DataIO')
    print(dataio)


    localdir, filenames, params = download_dataset(name='olfactory_bulb')
    dataio.set_data_source(type='RawData', filenames=filenames, **params)
    
    #with geometry
    channels = list(range(14))
    channel_groups = {0:{'channels':range(14), 'geometry' : { c: [0, i] for i, c in enumerate(channels) }}}
    dataio.set_channel_groups(channel_groups)
    
    #with no geometry
    channel_groups = {0:{'channels':range(4)}}
    dataio.set_channel_groups(channel_groups)
    
    # add one group
    dataio.add_one_channel_group(channels=range(4,8), chan_grp=5)
    
    
    channel_groups = {0:{'channels':range(14)}}
    dataio.set_channel_groups(channel_groups)
    
    for seg_num in range(dataio.nb_segment):
        for i_stop, sigs_chunk in dataio.iter_over_chunk(seg_num=seg_num, chunksize=1024):
            assert sigs_chunk.shape[0] == 1024
            assert sigs_chunk.shape[1] == 14
            #~ print(seg_num, i_stop, sigs_chunk.shape)
    
    
    #reopen existing
    dataio = DataIO(dirname='test_DataIO')
    print(dataio)
    
    
    for seg_num in range(dataio.nb_segment):
        #~ print('seg_num', seg_num)
        for i_stop, sigs_chunk in dataio.iter_over_chunk(seg_num=seg_num, chunksize=1024):
            assert sigs_chunk.shape[0] == 1024
            assert sigs_chunk.shape[1] == 14



def test_DataIO_probes():
    # initialze dataio
    if os.path.exists('test_DataIO'):
        shutil.rmtree('test_DataIO')
        
    dataio = DataIO(dirname='test_DataIO')
    print(dataio)


    localdir, filenames, params = download_dataset(name='olfactory_bulb')
    dataio.set_data_source(type='RawData', filenames=filenames,  **params)
    
    probe_filename = 'neuronexus/A4x8-5mm-100-400-413-A32.prb'
    dataio.download_probe(probe_filename)
    dataio.download_probe('neuronexus/A4x8-5mm-100-400-413-A32')
    
    #~ print(dataio.channel_groups)
    #~ print(dataio.channels)
    #~ print(dataio.info['probe_filename'])
    
    assert dataio.nb_channel(0) == 8
    assert probe_filename.split('/')[-1] == dataio.info['probe_filename']
    
    dataio = DataIO(dirname='test_DataIO')
    #~ print(dataio)
    


def test_dataio_catalogue():
    if os.path.exists('test_DataIO'):
        shutil.rmtree('test_DataIO')
    
    dataio = DataIO(dirname='test_DataIO')
    
    catalogue = {}
    catalogue['chan_grp'] = 0
    catalogue['centers0'] = np.ones((300, 12, 50))
    
    catalogue['n_left'] = -15
    catalogue['signal_preprocessor_params'] = {'highpass_freq' : 300.}
    
    dataio.save_catalogue(catalogue, name='test')
    
    c2 = dataio.load_catalogue(name='test', chan_grp=0)
    #~ print(c2)
    assert c2['n_left'] == -15
    assert np.all(c2['centers0']==1)
    assert catalogue['signal_preprocessor_params']['highpass_freq'] == 300.
    



if __name__=='__main__':
    
    test_DataIO()
    test_DataIO_probes()
    test_dataio_catalogue()
    
    


"""
Some help function to compute in parallel processing at some stage:

  * CatalogueConstructor.run_signalprocessor = preprocessing + peak detection

Used only for offline computation.
This is usefull mainlly when the IO are slow.


"""
import time
import os
import loky

#~ import concurrent.futures.ThreadPoolExecutor

import numpy as np

from .dataio import DataIO
from .signalpreprocessor import signalpreprocessor_engines
from .peakdetector import get_peak_detector_class
from . import labelcodes


# TODO VERY IMPORTANT MOVE THIS
_dtype_peak = [('index', 'int64'), ('cluster_label', 'int64'), ('channel', 'int64'),  ('segment', 'int64'), ('extremum_amplitude', 'float64'),]



def signalprocessor_initializer(dirname, chan_grp_, seg_num_,
                internal_dtype, chunksize, pad_width_,
                signals_medians, signals_mads,
                signal_preprocessor_params,
                peak_detector_params):
    global dataio
    global chan_grp
    global seg_num
    global pad_width
    global signalpreprocessor
    global peakdetector
    
    dataio = DataIO(dirname)
    #~ print('signalprocessor_initializer', id(dataio))
    
    chan_grp = chan_grp_
    
    seg_num = seg_num_
    
    pad_width = pad_width_
    
    
    p = dict(signal_preprocessor_params)
    engine = p.pop('engine')
    SignalPreprocessor_class = signalpreprocessor_engines[engine]
    signalpreprocessor = SignalPreprocessor_class(dataio.sample_rate, dataio.nb_channel(chan_grp), chunksize, dataio.source_dtype)
    p['normalize'] = True
    p['signals_medians'] = signals_medians
    p['signals_mads'] = signals_mads
    signalpreprocessor.change_params(**p)

    
    p = dict(peak_detector_params)
    engine = p.pop('engine')
    method = p.pop('method')
    PeakDetector_class = get_peak_detector_class(method, engine)
    geometry = dataio.get_geometry(chan_grp)
    peakdetector = PeakDetector_class(dataio.sample_rate, dataio.nb_channel(chan_grp),
                                                    chunksize, internal_dtype, geometry)
    peakdetector.change_params(**p)


def read_process_write_one_chunk(args):
    i_start, i_stop = args
    # this read process and write one chunk
    global dataio
    global chan_grp
    global seg_num
    global pad_width
    global signalpreprocessor
    global peakdetector
    
    #~ print(i_start, i_stop, id(dataio))
    #~ print(dataio)
    
    # read chunk and pad
    sigs_chunk = dataio.get_signals_chunk(seg_num=seg_num, chan_grp=chan_grp,
                i_start=i_start, i_stop=i_stop,  signal_type='initial', pad_width=pad_width)
    #~ print('pad_width', pad_width)
    #~ print('read_process_write_one_chunk', i_start, i_stop, i_stop-i_start, sigs_chunk.shape)

    # process
    preprocessed_chunk = signalpreprocessor.process_buffer(sigs_chunk)
    #~ exit()
    
    
    
    # peak detection
    n_span = peakdetector.n_span
    assert n_span < pad_width
    chunk_peak = preprocessed_chunk[pad_width-n_span:-pad_width+n_span]
    time_ind_peaks, chan_ind_peaks, peak_val_peaks = peakdetector.process_buffer(chunk_peak)
    
    peaks = np.zeros(time_ind_peaks.size, dtype=_dtype_peak)
    peaks['index'] = time_ind_peaks -  n_span+ i_start
    peaks['segment'][:] = seg_num
    peaks['cluster_label'][:] = labelcodes.LABEL_NO_WAVEFORM
    if chan_ind_peaks is None:
        peaks['channel'][:] = -1
    else:
        peaks['channel'][:] = chan_ind_peaks
    if peak_val_peaks is None:
        peaks['extremum_amplitude'][:] = 0.
    else:
        peaks['extremum_amplitude'][:] = peak_val_peaks
    
    
    # remove the padding and write
    preprocessed_chunk = preprocessed_chunk[pad_width:-pad_width]
    dataio.set_signals_chunk(preprocessed_chunk, seg_num=seg_num, chan_grp=chan_grp,
                    i_start=i_start, i_stop=i_stop, signal_type='processed')
    
    
    return peaks
    

def run_parallel_read_process_write(cc, seg_num, length, n_jobs):
    
    chunksize = cc.info['chunksize']
    pad_width = cc.info['preprocessor']['pad_width']
    
    initargs=(cc.dataio.dirname, cc.chan_grp, seg_num, 
                    cc.internal_dtype, chunksize, pad_width,
                    cc.signals_medians, cc.signals_mads,
                    cc.info['preprocessor'],
                    cc.info['peak_detector'],
                    )
    
    
    
    if length is None:
        length = cc.dataio.get_segment_length(seg_num)
    num_chunk = length // chunksize
    chunk_slice = [(i*chunksize, (i+1)*chunksize) for i in range(num_chunk)]
    if length % chunksize > 0:
        chunk_slice.append((num_chunk*chunksize, length))

    if n_jobs < 0:
        n_jobs = os.cpu_count() + 1 - n_jobs

    if n_jobs > 1:
        n_jobs = min(n_jobs, len(chunk_slice))
        executor = loky.get_reusable_executor(
            max_workers=n_jobs, initializer=signalprocessor_initializer,
            initargs=initargs, context="loky", timeout=20)
        
        #~ concurrent.futures.ThreadPoolExecutor
        #~ executor = 
        
        
        
        all_peaks = executor.map(read_process_write_one_chunk, chunk_slice)
        for peaks in all_peaks:
            #~ print('peaks', peaks.size)
            cc.arrays.append_chunk('all_peaks',  peaks)
    else:
        signalprocessor_initializer(*initargs)
        
        for sl in chunk_slice:
            peaks = read_process_write_one_chunk(sl)
            #~ print(peaks)
            cc.arrays.append_chunk('all_peaks',  peaks)
    
    cc.dataio.flush_processed_signals(seg_num=seg_num, chan_grp=cc.chan_grp, processed_length=int(length))





"""
Very basic comand line for tridesclous


"""
import sys
import os
import argparse


import pyqtgraph as pg
import tridesclous as tdc
import tridesclous.gui as tdcgui

comand_list =[
    'mainwin',
    #~ 'makecatalogue',
    #~ 'runpeeler',
    'cataloguewin',
    'peelerwin',
    'init',
    'rt_demo',
    'rt_openephys',
]
txt_command_list = ', '.join(comand_list)


def open_mainwindow():
    app = pg.mkQApp()
    win = tdcgui.MainWindow()
    win.show()
    app.exec_()


def main():
    argv = sys.argv[1:]

    parser = argparse.ArgumentParser(description='tridesclous')
    parser.add_argument('command', help='command in [{}]'.format(txt_command_list), default='mainwin', nargs='?')
    
    parser.add_argument('-d', '--dirname', help='working directory', default=None)
    parser.add_argument('-c', '--chan_grp', type=int, help='channel group index', default=0)
    parser.add_argument('-p', '--prb_file', help='Probe file', default=None)
    
    
    args = parser.parse_args(argv)
    #~ print(sys.argv)
    #~ print(args)
    #~ print(args.command)
    
    command = args.command
    if not command in comand_list:
        print('command should be in [{}]'.format(txt_command_list))
        exit()
    
    dirname = args.dirname
    if dirname is None:
        dirname = os.getcwd()
    
    #~ print(command)
    
    if command in ['cataloguewin', 'peelerwin']:
        if not tdc.DataIO.check_initialized(dirname):
            print('{} is not initialized'.format(dirname))
            exit()
        dataio = tdc.DataIO(dirname=dirname)
        print(dataio)
    
    if command=='mainwin':
        open_mainwindow()
    
    #~ elif command=='makecatalogue':
        #~ pass
    
    #~ elif command=='runpeeler':
        #~ pass
        
    elif command=='cataloguewin':
        catalogueconstructor = tdc.CatalogueConstructor(dataio=dataio, chan_grp=args.chan_grp)
        app = pg.mkQApp()
        win = tdcgui.CatalogueWindow(catalogueconstructor)
        win.show()
        app.exec_()    
        
    elif command=='peelerwin':
        initial_catalogue = dataio.load_catalogue(chan_grp=args.chan_grp)
        app = pg.mkQApp()
        win = tdcgui.PeelerWindow(dataio=dataio, catalogue=initial_catalogue)
        win.show()
        app.exec_()
    
    elif command=='init':
        app = pg.mkQApp()
        win = tdcgui.InitializeDatasetWindow()
        win.show()
        app.exec_()
        
    elif command == 'rt_demo':
        from tridesclous.online import start_online_pyacq_buffer_demo
        start_online_pyacq_buffer_demo()
    
    elif command == 'rt_openephys':
        from tridesclous.online import start_online_openephys
        start_online_openephys(prb_filename=args.prb_file)
        

if __name__ =='__main__':
    open_mainwindow()




from tridesclous.dip import diptest
import numpy as np

import time


def test_diptest():
    
    n = 4000

    data = np.concatenate([np.random.randn(n)-3, np.random.randn(n)+3])
    
    t1 = time.perf_counter()
    pval = diptest(data)
    t2 = time.perf_counter()
    print('diptest', t2-t1)    

    #~ assert pval < 0.001
    
    print(pval)
    
    
    data = np.concatenate([np.random.randn(n)-3])
    t1 = time.perf_counter()
    pval = diptest(data)
    t2 = time.perf_counter()
    print('diptest', t2-t1)    
    
    #~ assert pval > 0.2
    
    print(pval)




if __name__ == '__main__':
    
    test_diptest()


